# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import concurrent.futures
from itertools import repeat

from pyincore import AnalysisUtil, GeoUtil
from pyincore import BaseAnalysis, HazardService, FragilityService
from pyincore.models.dfr3curve import DFR3Curve


class EpfDamage(BaseAnalysis):
    """Computes electric power facility structural damage for an earthquake, tsunami, tornado, and hurricane hazards.

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    DEFAULT_LIQ_FRAGILITY_KEY = "pgd"
    DEFAULT_FRAGILITY_KEY = "pga"

    def __init__(self, incore_client):
        self.hazardsvc = HazardService(incore_client)
        self.fragilitysvc = FragilityService(incore_client)

        super(EpfDamage, self).__init__(incore_client)

    def run(self):
        """Executes electric power facility damage analysis."""
        epf_set = self.get_input_dataset("epfs").get_inventory_reader()

        # Get Fragility key
        fragility_key = self.get_parameter("fragility_key")
        if fragility_key is None:
            fragility_key = self.DEFAULT_FRAGILITY_KEY
            self.set_parameter("fragility_key", fragility_key)

        # get input hazard
        hazard, hazard_type, hazard_dataset_id = self.create_hazard_object_from_input_params()

        # Hazard Uncertainty
        use_hazard_uncertainty = False
        if self.get_parameter("use_hazard_uncertainty") is not None:
            use_hazard_uncertainty = self.get_parameter("use_hazard_uncertainty")

        if use_hazard_uncertainty:
            raise ValueError("Uncertainty is not implemented yet.")

        user_defined_cpu = 1

        if not self.get_parameter("num_cpu") is None and self.get_parameter("num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        num_workers = AnalysisUtil.determine_parallelism_locally(self, len(epf_set), user_defined_cpu)

        avg_bulk_input_size = int(len(epf_set) / num_workers)
        inventory_args = []
        count = 0
        inventory_list = list(epf_set)
        while count < len(inventory_list):
            inventory_args.append(inventory_list[count:count + avg_bulk_input_size])
            count += avg_bulk_input_size

        (ds_results, damage_results) = self.epf_damage_concurrent_future(self.epf_damage_analysis_bulk_input,
                                                                         num_workers,
                                                                         inventory_args,
                                                                         repeat(hazard),
                                                                         repeat(hazard_type),
                                                                         repeat(hazard_dataset_id))

        self.set_result_csv_data("result", ds_results, name=self.get_parameter("result_name"))
        self.set_result_json_data("metadata", damage_results,
                                  name=self.get_parameter("result_name") + "_additional_info")

        return True

    def epf_damage_concurrent_future(self, function_name, num_workers, *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            num_workers (int): Maximum number workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            list: A list of ordered dictionaries with epf damage values and other data/metadata.

        """

        output_ds = []
        output_dmg = []
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_workers) as executor:
            for ret1, ret2 in executor.map(function_name, *args):
                output_ds.extend(ret1)
                output_dmg.extend(ret2)

        return output_ds, output_dmg

    def epf_damage_analysis_bulk_input(self, epfs, hazard, hazard_type, hazard_dataset_id):
        """Run analysis for multiple epfs.

        Args:
            epfs (list): Multiple epfs from input inventory set.
            hazard (obj): A hazard object.
            hazard_type (str): A type of hazard exposure (earthquake, tsunami, tornado, or hurricane).
            hazard_dataset_id (str): An id of the hazard exposure.

        Returns:
            list: A list of ordered dictionaries with epf damage values and other data/metadata.

        """

        # Setting Local Variables
        use_liquefaction = False
        liquefaction_available = False
        fragility_sets_liq = None
        geology_dataset_id = None
        liq_demand_types = None
        liq_demand_units = None
        liq_hazard_vals = None
        liquefaction_prob = None

        fragility_key = self.get_parameter("fragility_key")

        fragility_set = self.fragilitysvc.match_inventory(self.get_input_dataset("dfr3_mapping_set"), epfs,
                                                          fragility_key)

        if hazard_type == "earthquake":
            liquefaction_fragility_key = self.get_parameter("liquefaction_fragility_key")
            if self.get_parameter("use_liquefaction") is True:
                if liquefaction_fragility_key is None:
                    liquefaction_fragility_key = self.DEFAULT_LIQ_FRAGILITY_KEY

                use_liquefaction = self.get_parameter("use_liquefaction")

                # Obtain the geology dataset
                geology_dataset_id = self.get_parameter("liquefaction_geology_dataset_id")

                if geology_dataset_id is not None:
                    fragility_sets_liq = self.fragilitysvc.match_inventory(
                        self.get_input_dataset("dfr3_mapping_set"), epfs, liquefaction_fragility_key)

                    if fragility_sets_liq is not None:
                        liquefaction_available = True

        values_payload = []
        values_payload_liq = []
        unmapped_epfs = []
        mapped_epfs = []
        for epf in epfs:
            epf_id = epf["id"]
            if epf_id in fragility_set:
                location = GeoUtil.get_location(epf)
                loc = str(location.y) + "," + str(location.x)
                demands = fragility_set[epf_id].demand_types
                units = fragility_set[epf_id].demand_units
                value = {
                    "demands": demands,
                    "units": units,
                    "loc": loc
                }
                values_payload.append(value)
                mapped_epfs.append(epf)

                if liquefaction_available and epf["id"] in fragility_sets_liq:
                    fragility_set_liq = fragility_sets_liq[epf["id"]]
                    demands_liq = fragility_set_liq.demand_types
                    units_liq = fragility_set_liq.demand_units
                    value_liq = {
                        "demands": demands_liq,
                        "units": units_liq,
                        "loc": loc
                    }
                    values_payload_liq.append(value_liq)
            else:
                unmapped_epfs.append(epf)

        hazard_vals = hazard.read_hazard_values(values_payload, self.hazardsvc)

        liquefaction_resp = None
        if liquefaction_available:
            liquefaction_resp = self.hazardsvc.post_liquefaction_values(hazard_dataset_id,
                                                                        geology_dataset_id,
                                                                        values_payload_liq)

        ds_results = []
        damage_results = []

        i = 0
        for epf in mapped_epfs:
            ds_result = dict()
            damage_result = dict()
            selected_fragility_set = fragility_set[epf["id"]]

            if isinstance(selected_fragility_set.fragility_curves[0], DFR3Curve):
                hazard_val = AnalysisUtil.update_precision_of_lists(hazard_vals[i]["hazardValues"])
                input_demand_types = hazard_vals[i]["demands"]
                input_demand_units = hazard_vals[i]["units"]

                hval_dict = dict()
                j = 0
                for d in selected_fragility_set.demand_types:
                    hval_dict[d] = hazard_val[j]
                    j += 1

                epf_args = selected_fragility_set.construct_expression_args_from_inventory(epf)
                limit_states = selected_fragility_set.calculate_limit_state(hval_dict,
                                                                            inventory_type='electric_facility',
                                                                            **epf_args)

                if liquefaction_resp is not None:
                    fragility_set_liq = fragility_sets_liq[epf["id"]]

                    if isinstance(fragility_set_liq.fragility_curves[0], DFR3Curve):
                        liq_hazard_vals = AnalysisUtil.update_precision_of_lists(liquefaction_resp[i]["pgdValues"])
                        liq_demand_types = liquefaction_resp[i]["demands"]
                        liq_demand_units = liquefaction_resp[i]["units"]
                        liquefaction_prob = liquefaction_resp[i]['liqProbability']

                        hval_dict_liq = dict()

                        for j, d in enumerate(fragility_set_liq.demand_types):
                            hval_dict_liq[d] = liq_hazard_vals[j]

                        facility_liq_args = fragility_set_liq.construct_expression_args_from_inventory(epf)
                        pgd_limit_states = \
                            fragility_set_liq.calculate_limit_state(
                                hval_dict_liq, inventory_type="electric_facility",
                                **facility_liq_args)
                    else:
                        raise ValueError("One of the fragilities is in deprecated format. "
                                         "This should not happen If you are seeing this please report the issue.")

                    limit_states = AnalysisUtil.adjust_limit_states_for_pgd(limit_states, pgd_limit_states)

                dmg_interval = selected_fragility_set.calculate_damage_interval(
                    limit_states, hazard_type=hazard_type, inventory_type='electric_facility')
            else:
                raise ValueError("One of the fragilities is in deprecated format. This should not happen. If you are "
                                 "seeing this please report the issue.")

            ds_result["guid"] = epf["properties"]["guid"]
            ds_result.update(limit_states)
            ds_result.update(dmg_interval)
            ds_result['haz_expose'] = AnalysisUtil.get_exposure_from_hazard_values(hazard_val, hazard_type)

            damage_result['guid'] = epf['properties']['guid']
            damage_result['fragility_id'] = selected_fragility_set.id
            damage_result["demandtypes"] = input_demand_types
            damage_result["demandunits"] = input_demand_units
            damage_result["hazardtype"] = hazard_type
            damage_result["hazardvals"] = hazard_val

            if hazard_type == "earthquake" and use_liquefaction is True:
                if liquefaction_available:
                    damage_result['liq_fragility_id'] = fragility_sets_liq[epf["id"]].id
                    damage_result['liqdemandtypes'] = liq_demand_types
                    damage_result['liqdemandunits'] = liq_demand_units
                    damage_result['liqhazval'] = liq_hazard_vals
                    damage_result['liqprobability'] = liquefaction_prob
                else:
                    damage_result['liq_fragility_id'] = None
                    damage_result['liqdemandtypes'] = None
                    damage_result['liqdemandunits'] = None
                    damage_result['liqhazval'] = None
                    damage_result['liqprobability'] = None

            ds_results.append(ds_result)
            damage_results.append(damage_result)

            i += 1

        #############################################################

        # unmapped
        for epf in unmapped_epfs:
            ds_result = dict()
            damage_result = dict()
            ds_result['guid'] = epf['properties']['guid']
            damage_result['guid'] = epf['properties']['guid']
            damage_result['fragility_id'] = None
            damage_result["demandtypes"] = None
            damage_result['demandunits'] = None
            damage_result["hazardtype"] = None
            damage_result['hazardval'] = None
            if hazard_type == "earthquake" and use_liquefaction is True:
                damage_result['liq_fragility_id'] = None
                damage_result['liqdemandtypes'] = None
                damage_result['liqdemandunits'] = None
                damage_result['liqhazval'] = None
                damage_result['liqprobability'] = None

            ds_results.append(ds_result)
            damage_results.append(damage_result)

        return ds_results, damage_results

    def get_spec(self):
        """Get specifications of the epf damage analysis.

        Returns:
            obj: A JSON object of specifications of the epf damage analysis.

        """
        return {
            'name': 'epf-damage',
            'description': 'Electric Power Facility damage analysis.',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'A name of the resulting dataset',
                    'type': str
                },
                {
                    'id': 'hazard_type',
                    'required': False,
                    'description': 'Hazard type (e.g. earthquake).',
                    'type': str
                },
                {
                    'id': 'hazard_id',
                    'required': False,
                    'description': 'Hazard ID which defines the particular hazard (e.g. New madrid earthquake '
                                   'using Atkinson Boore 1995).',
                    'type': str
                },
                {
                    'id': 'fragility_key',
                    'required': False,
                    'description': 'Fragility key to use in mapping dataset ()',
                    'type': str
                },
                {
                    'id': 'liquefaction_fragility_key',
                    'required': False,
                    'description': 'Fragility key to use in liquefaction mapping dataset',
                    'type': str
                },
                {
                    'id': 'use_liquefaction',
                    'required': False,
                    'description': 'Use a ground liquifacition to modify damage interval.',
                    'type': bool
                },
                {
                    'id': 'liquefaction_geology_dataset_id',
                    'required': False,
                    'description': 'Liquefaction geology/susceptibility dataset id. '
                                   'If not provided, liquefaction will be ignored',
                    'type': str
                },
                {
                    'id': 'use_hazard_uncertainty',
                    'required': False,
                    'description': 'Use hazard uncertainty',
                    'type': bool
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request.',
                    'type': int
                },
            ],
            'input_hazards': [
                {
                    'id': 'hazard',
                    'required': False,
                    'description': 'Hazard object',
                    'type': ["earthquake", "tornado", "hurricane", "flood", "tsunami"]
                },
            ],
            'input_datasets': [
                {
                    'id': 'epfs',
                    'required': True,
                    'description': 'Electric Power Facility Inventory',
                    'type': ['incore:epf',
                             'ergo:epf',
                             'incore:epfVer2'
                             ],
                },
                {
                    'id': 'dfr3_mapping_set',
                    'required': True,
                    'description': 'DFR3 Mapping Set Object',
                    'type': ['incore:dfr3MappingSet'],
                }
            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'epfs',
                    'description': 'A csv file with limit state probabilities and damage states '
                                   'for each electric power facility',
                    'type': 'incore:epfDamageVer3'
                },
                {
                    'id': 'metadata',
                    'parent_type': 'epfs',
                    'description': 'additional metadata in json file about applied hazard value and '
                                   'fragility',
                    'type': 'incore:epfDamageSupplement'
                }
            ]
        }



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

class HousingRecoveryUtil:
    BASEYEAR = 2008

    DMG_YEARS = [-1, 0, 1, 2, 3, 4, 5, 6]

    # The models presented in this section are based on models developed in Hamideh et al. (2018).
    # The modifications from the original Galveston models include (1) combining non-Hispanic Black
    # and Hispanic block group percentages into a minority block group percentage variable,
    # and (2) having both parcel and block group level intercepts. The original model only had parcel level
    # intercepts. Considering that multiple levels of data are used (parcel and block group), it is appropriate
    # to have both levels of intercepts.

    # Primary housing market (PHM) model coefficients
    B_PHM_intercept = 11.407594

    # Year indicator dummy variables
    B_PHM_year = {}
    B_PHM_year[-1] = 0.000000  # year -1, tax assessment immediately before disaster
    B_PHM_year[0] = 0.263500  # year  0, tax assessment immediately after disaster, damage year
    B_PHM_year[1] = 0.147208  # year +1
    B_PHM_year[2] = 0.110004  # year +2
    B_PHM_year[3] = 0.122228  # year +3
    B_PHM_year[4] = 0.102886  # year +4
    B_PHM_year[5] = 0.102806  # year +5
    B_PHM_year[6] = 0.276548  # year +6

    # House Characteristics
    B_PHM_age = -0.030273
    B_PHM_sqm = 0.005596

    # Damage and year dummy interactions
    B_PHM_dmg_year = {}
    B_PHM_dmg_year[-1] = 0.000000
    B_PHM_dmg_year[0] = -0.037458  # base effect
    B_PHM_dmg_year[1] = 0.009310 + B_PHM_dmg_year[0]
    B_PHM_dmg_year[2] = 0.009074 + B_PHM_dmg_year[0]
    B_PHM_dmg_year[3] = 0.010340 + B_PHM_dmg_year[0]
    B_PHM_dmg_year[4] = 0.011293 + B_PHM_dmg_year[0]
    B_PHM_dmg_year[5] = 0.012178 + B_PHM_dmg_year[0]
    B_PHM_dmg_year[6] = 0.012188 + B_PHM_dmg_year[0]

    # Owner-occupied and year dummy interactions
    B_PHM_own_year = {}
    B_PHM_own_year[-1] = 0.017153   # base effect
    B_PHM_own_year[0] = 0.129077 + B_PHM_own_year[-1]
    B_PHM_own_year[1] = 0.188217 + B_PHM_own_year[-1]
    B_PHM_own_year[2] = 0.235435 + B_PHM_own_year[-1]
    B_PHM_own_year[3] = 0.246437 + B_PHM_own_year[-1]
    B_PHM_own_year[4] = 0.261220 + B_PHM_own_year[-1]
    B_PHM_own_year[5] = 0.281015 + B_PHM_own_year[-1]
    B_PHM_own_year[6] = 0.265465 + B_PHM_own_year[-1]

    # Median household income and year dummy interactions
    B_PHM_inc_year = {}
    B_PHM_inc_year[-1] = 0.002724   # base effect
    B_PHM_inc_year[0] = 0.001190 + B_PHM_inc_year[-1]
    B_PHM_inc_year[1] = 0.001480 + B_PHM_inc_year[-1]
    B_PHM_inc_year[2] = 0.001746 + B_PHM_inc_year[-1]
    B_PHM_inc_year[3] = 0.001494 + B_PHM_inc_year[-1]
    B_PHM_inc_year[4] = 0.001480 + B_PHM_inc_year[-1]
    B_PHM_inc_year[5] = 0.001849 + B_PHM_inc_year[-1]
    B_PHM_inc_year[6] = 0.001400 + B_PHM_inc_year[-1]

    # Block Group percent Minority and year dummy interactions
    B_PHM_min_year = {}
    B_PHM_min_year[-1] = -0.004783   # base effect
    B_PHM_min_year[0] = 0.005609 + B_PHM_min_year[-1]
    B_PHM_min_year[1] = 0.007343 + B_PHM_min_year[-1]
    B_PHM_min_year[2] = 0.007459 + B_PHM_min_year[-1]
    B_PHM_min_year[3] = 0.006421 + B_PHM_min_year[-1]
    B_PHM_min_year[4] = 0.006308 + B_PHM_min_year[-1]
    B_PHM_min_year[5] = 0.005770 + B_PHM_min_year[-1]
    B_PHM_min_year[6] = 0.005620 + B_PHM_min_year[-1]

    # Seasonal/Vacation Housing Market (SVHM) model coefficients
    B_SVHM_intercept = 11.125636

    # Year indicator dummy variables
    B_SVHM_year = {}
    B_SVHM_year[-1] = 0.000000  # year -1, tax assessment immediately before disaster
    B_SVHM_year[0] = 1.489008  # year  0, tax assessment immediately after disaster, damage year
    B_SVHM_year[1] = 1.858770  # year +1
    B_SVHM_year[2] = 2.163492  # year +2
    B_SVHM_year[3] = 2.071690  # year +3
    B_SVHM_year[4] = 2.110245  # year +4
    B_SVHM_year[5] = 2.208577  # year +5
    B_SVHM_year[6] = 2.273867  # year +6

    # House Characteristics
    B_SVHM_age = -0.014260
    B_SVHM_sqm = 0.005505

    # Damage and year dummy interactions
    B_SVHM_dmg_year = {}
    B_SVHM_dmg_year[-1] = 0.000000
    B_SVHM_dmg_year[0] = -0.051913  # base effect
    B_SVHM_dmg_year[1] = 0.001504 + B_SVHM_dmg_year[0]
    B_SVHM_dmg_year[2] = -0.004172 + B_SVHM_dmg_year[0]
    B_SVHM_dmg_year[3] = -0.002016 + B_SVHM_dmg_year[0]
    B_SVHM_dmg_year[4] = 0.001687 + B_SVHM_dmg_year[0]
    B_SVHM_dmg_year[5] = 0.000876 + B_SVHM_dmg_year[0]
    B_SVHM_dmg_year[6] = 0.001129 + B_SVHM_dmg_year[0]

    # Owner-occupied and year dummy interactions
    B_SVHM_own_year = {}
    B_SVHM_own_year[-1] = -0.017167   # base effect
    B_SVHM_own_year[0] = 0.043263 + B_SVHM_own_year[-1]
    B_SVHM_own_year[1] = 0.003315 + B_SVHM_own_year[-1]
    B_SVHM_own_year[2] = 0.034372 + B_SVHM_own_year[-1]
    B_SVHM_own_year[3] = -0.014929 + B_SVHM_own_year[-1]
    B_SVHM_own_year[4] = -0.021078 + B_SVHM_own_year[-1]
    B_SVHM_own_year[5] = 0.001161 + B_SVHM_own_year[-1]
    B_SVHM_own_year[6] = 0.017562 + B_SVHM_own_year[-1]

    # Median household income and year dummy interactions
    B_SVHM_inc_year = {}
    B_SVHM_inc_year[-1] = 0.003786   # base effect
    B_SVHM_inc_year[0] = -0.013662 + B_SVHM_inc_year[-1]
    B_SVHM_inc_year[1] = -0.017401 + B_SVHM_inc_year[-1]
    B_SVHM_inc_year[2] = -0.021541 + B_SVHM_inc_year[-1]
    B_SVHM_inc_year[3] = -0.019756 + B_SVHM_inc_year[-1]
    B_SVHM_inc_year[4] = -0.020789 + B_SVHM_inc_year[-1]
    B_SVHM_inc_year[5] = -0.021555 + B_SVHM_inc_year[-1]
    B_SVHM_inc_year[6] = -0.019781 + B_SVHM_inc_year[-1]



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
import collections
import traceback
import random
from pyincore import BaseAnalysis


class ExampleAnalysis(BaseAnalysis):
    """Example Analysis demonstrates how to use the base analysis class by loading in building data and computing some
    mock damage output and writing the result dataset
    """

    def run(self):
        """Executes building damage analysis."""
        # Building dataset
        bldg_set = self.get_input_dataset("buildings").get_inventory_reader()

        results = []
        # Iterate over the building dataset and compute damage for each
        for building in bldg_set:
            results.append(self.building_damage_analysis(building))

        # Create the result dataset
        self.set_result_csv_data("result", results, name=self.get_parameter("result_name"))

        return True

    def building_damage_analysis(self, building):
        """Calculates building damage results for a single building.

        Args:
            building (obj): A JSON mapping of a geometric object from the inventory: current building.

        Returns:
            OrderedDict: A dictionary with building damage values and other data/metadata.

        """
        try:
            # Uncomment this to print building attributes
            # print(building)
            bldg_results = collections.OrderedDict()

            mean_damage = collections.OrderedDict()
            mean_damage['meandamage'] = random.uniform(0.0, 1.0)

            # Add building global id so damage can be linked to building attributes
            bldg_results['guid'] = building['properties']['guid']
            # Damage result
            bldg_results.update(mean_damage)

            return bldg_results

        except Exception as e:
            # This prints the type, value and stacktrace of error being handled.
            traceback.print_exc()
            print()
            raise e

    def get_spec(self):
        """Get specifications of the building damage analysis.

        Returns:
            obj: A JSON object of specifications of the building damage analysis.

        """
        return {
            'name': 'mock-building-damage',
            'description': 'mock-building damage analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
            ],
            'input_datasets': [
                {
                    'id': 'buildings',
                    'required': True,
                    'description': 'Building Inventory',
                    'type': ['ergo:buildingInventoryVer4', 'ergo:buildingInventoryVer5'],
                },

            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'buildings',
                    'description': 'CSV file of building structural damage',
                    'type': 'ergo:buildingDamageVer4'
                }
            ]
        }



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import concurrent.futures

from pyincore import AnalysisUtil
from pyincore import BaseAnalysis


class EpfRepairCost(BaseAnalysis):
    """Computes electric power facility repair cost.

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        super(EpfRepairCost, self).__init__(incore_client)

    def run(self):
        """Executes electric power facility repair cost analysis."""

        epf_df = self.get_input_dataset("epfs").get_dataframe_from_shapefile()
        sample_damage_states_df = self.get_input_dataset("sample_damage_states").get_dataframe_from_csv()
        replacement_cost = self.get_input_dataset("replacement_cost").get_dataframe_from_csv()

        # join damage state, replacement cost, with original inventory
        epf_df = epf_df.merge(sample_damage_states_df, on="guid")
        epf_df = epf_df.merge(replacement_cost, on="guid")
        epf_set = epf_df.to_dict(orient="records")

        user_defined_cpu = 1
        if not self.get_parameter("num_cpu") is None and self.get_parameter("num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        num_workers = AnalysisUtil.determine_parallelism_locally(self, len(epf_set), user_defined_cpu)

        avg_bulk_input_size = int(len(epf_set) / num_workers)
        inventory_args = []
        count = 0
        inventory_list = list(epf_set)
        while count < len(inventory_list):
            inventory_args.append(inventory_list[count:count + avg_bulk_input_size])
            count += avg_bulk_input_size

        repair_costs = self.epf_repair_cost_concurrent_future(self.epf_repair_cost_bulk_input, num_workers,
                                                              inventory_args)
        self.set_result_csv_data("result", repair_costs, name=self.get_parameter("result_name") + "_repair_cost")

        return True

    def epf_repair_cost_concurrent_future(self, function_name, num_workers, *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            num_workers (int): Maximum number workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            list: A list of ordered dictionaries with epf repair cost values and other data/metadata.

        """

        output = []
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_workers) as executor:
            for ret1 in executor.map(function_name, *args):
                output.extend(ret1)

        return output

    def epf_repair_cost_bulk_input(self, epfs):
        """Run analysis for multiple epfs.

        Args:
            epfs (list): Multiple epfs from input inventory set.

        Returns:
            list: A list of ordered dictionaries with epf repair cost values and other data/metadata.

        """
        # read in the damage ratio tables
        epf_dmg_ratios_csv = self.get_input_dataset("epf_dmg_ratios").get_csv_reader()
        dmg_ratio_tbl = AnalysisUtil.get_csv_table_rows(epf_dmg_ratios_csv, ignore_first_row=False)

        repair_costs = []

        for epf in epfs:
            rc = dict()
            rc["guid"] = epf["guid"]
            epf_type = epf["utilfcltyc"]

            sample_damage_states = epf["sample_damage_states"].split(",")
            repair_cost = ["0"] * len(sample_damage_states)
            for n, ds in enumerate(sample_damage_states):
                for dmg_ratio_row in dmg_ratio_tbl:
                    # use "in" instead of "==" since some inventory has pending number (e.g. EDC2)
                    if dmg_ratio_row["Inventory Type"] in epf_type and dmg_ratio_row["Damage State"] == ds:
                        dr = float(dmg_ratio_row["Best Mean Damage Ratio"])
                        repair_cost[n] = str(epf["replacement_cost"] * dr)

            rc["budget"] = ','.join(repair_cost)
            rc["repaircost"] = ','.join(repair_cost)

            repair_costs.append(rc)

        return repair_costs

    def get_spec(self):
        """Get specifications of the epf repair cost analysis.

        Returns:
            obj: A JSON object of specifications of the epf repair cost analysis.

        """
        return {
            "name": "epf-repair-cost",
            "description": "Electric Power Facility repair cost analysis.",
            "input_parameters": [
                {
                    "id": "result_name",
                    "required": True,
                    "description": "A name of the resulting dataset",
                    "type": str
                },
                {
                    "id": "num_cpu",
                    "required": False,
                    "description": "If using parallel execution, the number of cpus to request.",
                    "type": int
                },
            ],
            "input_datasets": [
                {
                    "id": "epfs",
                    "required": True,
                    "description": "Electric Power Facility Inventory",
                    "type": ["incore:epf",
                             "ergo:epf",
                             "incore:epfVer2"
                             ],
                },
                {
                    "id": "replacement_cost",
                    "required": True,
                    "description": "Repair cost of the node in the complete damage state (= Replacement cost)",
                    "type": ["incore:replacementCost"],
                },
                {
                    "id": "sample_damage_states",
                    "required": True,
                    "description": "sample damage states from Monte Carlo Simulation",
                    "type": ["incore:sampleDamageState"]
                },
                {
                    "id": "epf_dmg_ratios",
                    "required": True,
                    "description": "Damage Ratios table",
                    "type": ["incore:epfDamageRatios"]
                },
            ],
            "output_datasets": [
                {
                    "id": "result",
                    "parent_type": "epfs",
                    "description": "A csv file with repair cost for each electric power facility",
                    "type": "incore:repairCost"
                }
            ]
        }



# Copyright (c) 2021 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import json
import pandas as pd
import geopandas as gpd
from shapely import wkt


class PopDislOutputProcess:
    """This class converts csv results outputs of Population dislocation analysis to json format and shapefiles.

    Args:
        pop_disl_result (obj): IN-CORE dataset for Joplin Population Dislocation (PD) results.
        pop_disl_result_path (obj): A fallback for the case that Joplin PD object is not provided.
            For example a user wants to directly pass in csv files, a path to PD results.
        filter_name (str): A string to filter data by name, default empty. Example: filter_name="Joplin" for Joplin
            inventory, other is Duquesne etc. Name must be valid.
        filter_guid (bool): A flag to filter all data, default True counts only Joplin buildings.
        vacant_disl (bool): A flag to include vacant (Vacant for tenure) dislocation

    """
    HUPD_CATEGORIES = ["household_characteristics",
                       "household_dislocated",
                       "total_households",
                       "percent_household_dislocated",
                       "population_dislocated",
                       "total_population",
                       "percent_population_dislocated"
                       ]

    def __init__(self, pop_disl_result, pop_disl_result_path=None,
                 filter_name=None, filter_guid=True, vacant_disl=True):
        if pop_disl_result_path:
            pd_result = pd.read_csv(pop_disl_result_path, low_memory=False)
        else:
            pd_result = pop_disl_result.get_dataframe_from_csv(low_memory=False)
        pd_result["geometry"] = pd_result["geometry"].apply(wkt.loads)

        # keep only inventory with guid; filter for Joplin since only Joplin inventory has guids
        if filter_guid:
            if filter_name:
                pd_result_flag = pd_result[(pd_result["guid"].notnull()) &
                                           (pd_result["numprec"].notnull()) &
                                           (pd_result["plcname10"] == filter_name)]
                # only keep guid, place and dislocated
                pd_result_shp = pd_result[(pd_result["dislocated"]) &
                                          (pd_result["guid"].notnull()) &
                                          (pd_result["numprec"].notnull()) &
                                          (pd_result["plcname10"] == filter_name)]
            else:
                pd_result_flag = pd_result[(pd_result["guid"].notnull()) &
                                           (pd_result["numprec"].notnull())]
                # only keep guid and dislocated
                pd_result_shp = pd_result[(pd_result["dislocated"]) &
                                          (pd_result["guid"].notnull()) &
                                          (pd_result["numprec"].notnull())]
        else:
            if filter_name:
                pd_result_flag = pd_result[(pd_result["numprec"].notnull()) &
                                           (pd_result["plcname10"] == filter_name)]
                # only keep guid, place and dislocated
                pd_result_shp = pd_result[(pd_result["dislocated"]) &
                                          (pd_result["numprec"].notnull()) &
                                          (pd_result["plcname10"] == filter_name)]
            else:
                pd_result_flag = pd_result[(pd_result["numprec"].notnull())]
                # only keep guid and dislocated
                pd_result_shp = pd_result[(pd_result["dislocated"]) &
                                          (pd_result["numprec"].notnull())]
        self.vacant_disl = vacant_disl
        self.pop_disl_result = pd_result_flag
        self.pop_disl_result_shp = pd_result_shp

    def get_heatmap_shp(self, filename="pop-disl-numprec.shp"):
        """ Convert and filter population dislocation output to shapefile that contains only guid and numprec columns

            Args:
                filename (str): Path and name to save shapefile output file in. E.g "heatmap.shp"

            Returns:
                str: full path and filename of the shapefile

        """
        df = self.pop_disl_result_shp

        # save as shapefile
        gdf = gpd.GeoDataFrame(df, crs='epsg:4326')
        gdf = gdf[["guid", "numprec", "geometry"]]
        gdf.to_file(filename)

        return filename

    def pd_by_race(self, filename_json=None):
        """ Calculate race results from the output files of the Joplin Population Dislocation analysis
        and convert the results to json format.
        [
            {"household_characteristics": "Not Hispanic/White",
             "household_dislocated": 1521,
             "total_households": 18507,
             "%_household_dislocated": 7.3,
             "population_dislocated",
             "total_population",
             "%_population_dislocated"
            },{"household_characteristics": "Not Hispanic/Black",..,..},{},{},
            {"No race Ethnicity Data"},{"Total"}
        ]

        Args:
            filename_json (str): Path and name to save json output file in. E.g "pd_race_count.json"

        Returns:
            obj: PD total count by race. A JSON of the hua and population dislocation race results by category.

        """
        # Race categories
        # The numbering follows the Community description notebook
        # 0 - Vacant HU No Race Ethnicity Data, 1 - Not Hispanic/White, 2 - Not Hispanic/Black
        # 3 - Not Hispanic/Other race, 4 - Hispanic, 5 - No Race or Ethnicity Data
        race_categories = ["Vacant HU No Race or Ethnicity Data",
                           "Not Hispanic/White",
                           "Not Hispanic/Black",
                           "Not Hispanic/Other Race",
                           "Hispanic",
                           "No Race or Ethnicity Data",
                           "Total"]

        huapd = self.pop_disl_result
        # Allocated by race and ethnicity
        huapd["hua_re"] = "0"
        huapd.loc[(huapd["race"] == 1) & (huapd["hispan"] == 0), "hua_re"] = "1"
        huapd.loc[(huapd["race"] == 2) & (huapd["hispan"] == 0), "hua_re"] = "2"
        huapd.loc[(huapd["race"].isin([3, 4, 5, 6, 7])) & (huapd["hispan"] == 0), "hua_re"] = "3"
        huapd.loc[(huapd["hispan"] == 1), "hua_re"] = "4"
        huapd.loc[(huapd["gqtype"] >= 1), "hua_re"] = "5"
        hua_vals = huapd["hua_re"].value_counts()

        hua_tot = []
        for i in range(len(race_categories) - 1):
            try:
                hua_tot.append(int(hua_vals[str(i)]))
            except Exception:
                hua_tot.append(0)
        hua_tot.append(int(sum(hua_tot)))

        pop_tot = []
        for i in range(len(race_categories) - 1):
            pop_tot.append(int(huapd["numprec"].where(huapd["hua_re"] == str(i)).sum()))
        pop_tot.append(int(sum(pop_tot)))

        # Dislocated by race and ethnicity
        huapd["hud_re"] = ""
        huapd.loc[huapd["dislocated"], "hud_re"] = "0"
        huapd.loc[(huapd["race"] == 1) & (huapd["hispan"] == 0) & huapd["dislocated"], "hud_re"] = "1"
        huapd.loc[(huapd["race"] == 2) & (huapd["hispan"] == 0) & huapd["dislocated"], "hud_re"] = "2"
        huapd.loc[(huapd["race"].isin([3, 4, 5, 6, 7])) & (huapd["hispan"] == 0) & huapd["dislocated"], "hud_re"] = "3"
        huapd.loc[(huapd["hispan"] == 1) & huapd["dislocated"], "hud_re"] = "4"
        huapd.loc[(huapd["gqtype"] >= 1) & huapd["dislocated"], "hud_re"] = "5"
        hud_vals = huapd["hud_re"].value_counts()

        hua_disl = []
        for i in range(len(race_categories) - 1):
            try:
                hua_disl.append(int(hud_vals[str(i)]))
            except Exception:
                hua_disl.append(0)
        hua_disl.append(int(sum(hua_disl)))

        pd_disl = []
        for i in range(len(race_categories) - 1):
            pd_disl.append(int(huapd["numprec"].where(huapd["hud_re"] == str(i)).sum()))
        pd_disl.append(int(sum(pd_disl)))

        pd_by_race_json = []
        for i in range(len(race_categories)):
            huapd_race = {}
            huapd_race[self.HUPD_CATEGORIES[0]] = race_categories[i]
            huapd_race[self.HUPD_CATEGORIES[1]] = hua_disl[i]
            huapd_race[self.HUPD_CATEGORIES[2]] = hua_tot[i]
            if hua_tot[i]:
                huapd_race[self.HUPD_CATEGORIES[3]] = 100 * (hua_disl[i] / hua_tot[i])
            else:
                huapd_race[self.HUPD_CATEGORIES[3]] = None
            huapd_race[self.HUPD_CATEGORIES[4]] = pd_disl[i]
            huapd_race[self.HUPD_CATEGORIES[5]] = pop_tot[i]
            if pop_tot[i]:
                huapd_race[self.HUPD_CATEGORIES[6]] = 100 * (pd_disl[i] / pop_tot[i])
            else:
                huapd_race[self.HUPD_CATEGORIES[6]] = None
            pd_by_race_json.append(huapd_race)
        # print(pd_by_race_json)

        if filename_json:
            with open(filename_json, "w") as outfile:
                json.dump(pd_by_race_json, outfile)
        # Serializing json
        return json.dumps(pd_by_race_json)

    def pd_by_income(self, filename_json=None):
        """ Calculate income results from the output files of the Joplin Population Dislocation analysis
        and convert the results to json format.
        [
            {"household_characteristics": "HH1 (less than $15,000)",
             "household_dislocated": 311,
             "total_households": 3252,
             "%_household_dislocated": 7.3,
             "population_dislocated": 311,
             "total_population": 3252,
             "%_population_dislocated"
             },
             {"HH2 ($15,000 to $35,000)",..,..,..,..},{},{},{},{},
             {"Unknown",..,..,..,..}
        ]

        Args:
            filename_json (str): Path and name to save json output file in. E.g "pd_income_count.json"

        Returns:
            obj: PD total count by income. A JSON of the hua and population dislocation income results by category.

        """
        income_categories = ["HH1 (less than $15,000)",
                             "HH2 ($15,000 to $35,000)",
                             "HH3 ($35,000 to $70,000)",
                             "HH4 ($70,000 to $120,000)",
                             "HH5 (More than $120,000)",
                             "Unknown",
                             "Total"]

        huapd = self.pop_disl_result
        # Allocated by income
        hua_tot = []
        for i in range(1, 6):
            alloc_inc = (huapd["hhinc"] == i).sum()
            hua_tot.append(int(alloc_inc))
        hua_tot.append(int(pd.isna(huapd["hhinc"]).sum()))
        hua_tot.append(int(sum(hua_tot)))

        pop_tot = []
        for i in range(1, 6):
            alloc = huapd["numprec"].where(huapd["hhinc"] == i).sum()
            pop_tot.append(int(alloc))
        pop_tot.append(int(huapd["numprec"].where(pd.isna(huapd["hhinc"])).sum()))
        pop_tot.append(int(sum(pop_tot)))

        # Dislocated by income
        hua_disl = []
        for i in range(1, 6):
            disl = huapd.loc[(huapd["hhinc"] == i) & huapd["dislocated"], ["dislocated"]].sum()
            hua_disl.append(int(disl))
        disl_unknown = huapd.loc[pd.isna(huapd["hhinc"]) & huapd["dislocated"], ["dislocated"]].sum()
        hua_disl.append(int(disl_unknown))
        hua_disl.append(int(sum(hua_disl)))

        pd_disl = []
        for i in range(1, 6):
            disl = huapd.loc[(huapd["hhinc"] == i) & huapd["dislocated"], ["numprec"]].sum()
            pd_disl.append(int(disl))
        disl_unknown = huapd.loc[pd.isna(huapd["hhinc"]) & huapd["dislocated"], ["numprec"]].sum()
        pd_disl.append(int(disl_unknown))
        pd_disl.append(int(sum(pd_disl)))

        pd_by_income_json = []
        for i in range(len(income_categories)):
            huapd_income = {}
            huapd_income[self.HUPD_CATEGORIES[0]] = income_categories[i]
            huapd_income[self.HUPD_CATEGORIES[1]] = hua_disl[i]
            huapd_income[self.HUPD_CATEGORIES[2]] = hua_tot[i]
            if hua_tot[i]:
                huapd_income[self.HUPD_CATEGORIES[3]] = 100 * (hua_disl[i] / hua_tot[i])
            else:
                huapd_income[self.HUPD_CATEGORIES[3]] = None
            huapd_income[self.HUPD_CATEGORIES[4]] = pd_disl[i]
            huapd_income[self.HUPD_CATEGORIES[5]] = pop_tot[i]
            if pop_tot[i]:
                huapd_income[self.HUPD_CATEGORIES[6]] = 100 * (pd_disl[i] / pop_tot[i])
            else:
                huapd_income[self.HUPD_CATEGORIES[6]] = None
            pd_by_income_json.append(huapd_income)
        # print(pd_by_income_json)

        if filename_json:
            with open(filename_json, "w") as outfile:
                json.dump(pd_by_income_json, outfile)
        # Serializing json
        return json.dumps(pd_by_income_json)

    def pd_by_tenure(self, filename_json=None):
        """ Calculate tenure results from the output files of the Joplin Population Dislocation analysis
        and convert the results to json format.
        [
            {"household_characteristics": "Owner occupied",
             "household_dislocated": 1018,
             "total_households": 11344,
             "%_household_dislocated": 7.3,
             "population_dislocated": 1018,
             "total_population": 11344,
             "%_population_dislocated"
            },
            {"household_characteristics": "Renter occupied",..,..,..,..},{},{},{},{},{},
            {"total",..,..,..,..}
        ]

        Args:
            filename_json (str): Path and name to save json output file in. E.g "pd_income_count.json"

        Returns:
            obj: PD total count by income. A JSON of the hua and population dislocation income results by category.

        """
        # Tenure categories
        # The numbering follows the Community description notebook
        # 0 - Vacant HU No Tenure Data, 1 - Owner occupied, 2 - Renter occupied,
        # 3 - Nursing facilities, 4 - Other group quarters, 5 - Vacant for rent
        # 6 - Vacant for sale, 7 - Vacant other
        tenure_categories = ["Owner occupied",
                             "Renter occupied",
                             "Nursing facilities",
                             "Other group quarters",
                             "Vacant for rent",
                             "Vacant for sale",
                             "Vacant other",
                             "Total"]

        huapd = self.pop_disl_result
        # Allocated by tenure
        huapd["hua_tnr"] = "0"
        huapd.loc[huapd["ownershp"] == 1.0, "hua_tnr"] = "1"
        huapd.loc[huapd["ownershp"] == 2.0, "hua_tnr"] = "2"
        huapd.loc[huapd["gqtype"] == 3, "hua_tnr"] = "3"
        huapd.loc[huapd["gqtype"].isin([1, 2, 4, 5, 6, 7, 8]), "hua_tnr"] = "4"
        huapd.loc[huapd["vacancy"].isin([1, 2]), "hua_tnr"] = "5"
        huapd.loc[huapd["vacancy"].isin([3, 4]), "hua_tnr"] = "6"
        huapd.loc[huapd["vacancy"].isin([5, 6, 7]), "hua_tnr"] = "7"
        hua_vals = huapd["hua_tnr"].value_counts()
        hua_tot = []
        for i in range(len(tenure_categories)):
            try:
                hua_tot.append(int(hua_vals[str(i)]))
            except Exception:
                hua_tot.append(0)
        hua_tot.append(int(sum(hua_tot[1:])))

        pop_tot = []
        for i in range(len(tenure_categories)):
            pop_tot.append(int(huapd["numprec"].where(huapd["hua_tnr"] == str(i)).sum()))
        pop_tot.append(int(sum(pop_tot[1:])))

        # Dislocated by tenure
        huapd["hud_tnr"] = ""
        huapd.loc[huapd["dislocated"], "hud_tnr"] = "0"
        huapd.loc[(huapd["ownershp"] == 1.0) & huapd["dislocated"], "hud_tnr"] = "1"
        huapd.loc[(huapd["ownershp"] == 2.0) & huapd["dislocated"], "hud_tnr"] = "2"
        huapd.loc[(huapd["gqtype"] == 3) & huapd["dislocated"], "hud_tnr"] = "3"
        huapd.loc[huapd["gqtype"].isin([1, 2, 4, 5, 6, 7, 8]) & huapd["dislocated"], "hud_tnr"] = "4"
        huapd.loc[huapd["vacancy"].isin([1, 2]) & huapd["dislocated"], "hud_tnr"] = "5"
        huapd.loc[huapd["vacancy"].isin([3, 4]) & huapd["dislocated"], "hud_tnr"] = "6"
        huapd.loc[huapd["vacancy"].isin([5, 6, 7]) & huapd["dislocated"], "hud_tnr"] = "7"
        hud_vals = huapd["hud_tnr"].value_counts()
        hua_disl = []
        for i in range(len(tenure_categories)):
            try:
                hua_disl.append(int(hud_vals[str(i)]))
            except Exception:
                hua_disl.append(0)
        # If vacant_disl is False the Vacant places do not dislocate (set to 0).
        for i in range(len(tenure_categories)):
            if not self.vacant_disl and "Vacant" in tenure_categories[i]:
                hua_disl[i + 1] = 0

        hua_disl.append(int(sum(hua_disl[1:])))

        pd_disl = []
        for i in range(len(tenure_categories)):
            pd_disl.append(int(huapd["numprec"].where(huapd["hud_tnr"] == str(i)).sum()))
        pd_disl.append(int(sum(pd_disl[1:])))

        pd_by_tenure_json = []
        for i in range(len(tenure_categories)):
            huapd_tenure = {}
            huapd_tenure[self.HUPD_CATEGORIES[0]] = tenure_categories[i]
            huapd_tenure[self.HUPD_CATEGORIES[1]] = hua_disl[i + 1]
            huapd_tenure[self.HUPD_CATEGORIES[2]] = hua_tot[i + 1]
            if hua_tot[i + 1]:
                huapd_tenure[self.HUPD_CATEGORIES[3]] = 100 * (hua_disl[i + 1] / hua_tot[i + 1])
            else:
                huapd_tenure[self.HUPD_CATEGORIES[3]] = None
            huapd_tenure[self.HUPD_CATEGORIES[4]] = pd_disl[i + 1]
            huapd_tenure[self.HUPD_CATEGORIES[5]] = pop_tot[i + 1]
            if pop_tot[i + 1]:
                huapd_tenure[self.HUPD_CATEGORIES[6]] = 100 * (pd_disl[i + 1] / pop_tot[i + 1])
            else:
                huapd_tenure[self.HUPD_CATEGORIES[6]] = None
            pd_by_tenure_json.append(huapd_tenure)
        # print(pd_by_tenure_json)

        if filename_json:
            with open(filename_json, "w") as outfile:
                json.dump(pd_by_tenure_json, outfile)
        # Serializing json
        return json.dumps(pd_by_tenure_json)

    def pd_by_housing(self, filename_json=None):
        """ Calculate housing results from the output files of the Joplin Population Dislocation analysis
        using huestimate column (huestimate = 1 is single family, huestimate > 1 means multi family house)
        and convert the results to json format.
        [
            {"household_characteristics": "Single Family",
             "household_dislocated": 1162,
             "total_households": 837,
             "%_household_dislocated": 7.3,
             "population_dislocated": 1162,
             "total_population": 837,
             "%_population_dislocated"
             },{},{"Total",..,..,..,..}
        ]

        Args:
            filename_json (str): Path and name to save json output file in. E.g "pd_housing_count.json"

        Returns:
            obj: PD total count by housing. A JSON of the hua and population dislocation housing results by category.

        """
        # Household categories
        # 0 - Vacant HU No Tenure Data, 1 - Single Family, 2 - Multi Family
        household_categories = ["Single Family",
                                "Multi Family",
                                "Total"]

        huapd = self.pop_disl_result
        # Allocated by housing
        huapd["hua_house"] = "0"
        huapd.loc[(huapd["huestimate"] == 1.0), "hua_house"] = "1"
        huapd.loc[(huapd["huestimate"] > 1.0), "hua_house"] = "2"
        hua_vals = huapd["hua_house"].value_counts()
        hua_tot = []
        for i in range(len(household_categories)):
            try:
                hua_tot.append(int(hua_vals[str(i)]))
            except Exception:
                hua_tot.append(0)
        hua_tot.append(int(sum(hua_tot[1:])))

        pop_tot = []
        for i in range(len(household_categories)):
            pop_tot.append(int(huapd["numprec"].where(huapd["hua_house"] == str(i)).sum()))
        pop_tot.append(int(sum(pop_tot[1:])))

        # Dislocated by household
        huapd["hud_house"] = ""
        huapd.loc[huapd["dislocated"], "hud_house"] = "0"
        huapd.loc[(huapd["huestimate"] == 1.0) & huapd["dislocated"], "hud_house"] = "1"
        huapd.loc[(huapd["huestimate"] > 1.0) & huapd["dislocated"], "hud_house"] = "2"
        hud_vals = huapd["hud_house"].value_counts()
        hua_disl = []
        for i in range(len(household_categories)):
            try:
                hua_disl.append(int(hud_vals[str(i)]))
            except Exception:
                hua_disl.append(0)
        hua_disl.append(int(sum(hua_disl[1:])))

        pd_disl = []
        for i in range(len(household_categories)):
            pd_disl.append(int(huapd["numprec"].where(huapd["hud_house"] == str(i)).sum()))
        pd_disl.append(int(sum(pd_disl[1:])))

        pd_by_housing_json = []
        for i in range(len(household_categories)):
            huapd_household = {}
            huapd_household[self.HUPD_CATEGORIES[0]] = household_categories[i]
            huapd_household[self.HUPD_CATEGORIES[1]] = hua_disl[i + 1]
            huapd_household[self.HUPD_CATEGORIES[2]] = hua_tot[i + 1]
            if hua_tot[i + 1]:
                huapd_household[self.HUPD_CATEGORIES[3]] = 100 * (hua_disl[i + 1] / hua_tot[i + 1])
            else:
                huapd_household[self.HUPD_CATEGORIES[3]] = None
            huapd_household[self.HUPD_CATEGORIES[4]] = pd_disl[i + 1]
            huapd_household[self.HUPD_CATEGORIES[5]] = pop_tot[i + 1]
            if pop_tot[i + 1]:
                huapd_household[self.HUPD_CATEGORIES[6]] = 100 * (pd_disl[i + 1] / pop_tot[i + 1])
            else:
                huapd_household[self.HUPD_CATEGORIES[6]] = None
            pd_by_housing_json.append(huapd_household)
        # print(pd_by_housing_json)

        if filename_json:
            with open(filename_json, "w") as outfile:
                json.dump(pd_by_housing_json, outfile)
        return json.dumps(pd_by_housing_json)

    def pd_total(self, filename_json=None):
        """ Calculate total results from the output files of the Joplin Population Dislocation analysis
        and convert the results to json format.
        {   "household_dislocated": {
                "dislocated": {
                    "number": 1999,
                    "percentage": 0.085
                }, "not_dislocated": {}, "total": {}
            },"population_dislocated": {"dislocated": {},"not_dislocated": {}, "total": {}}
        }

        Args:
            filename_json (str): Path and name to save json output file in. E.g "pd_total_count.json"

        Returns:
            obj: PD total count. A JSON of the hua and population dislocation total results by category.

        """
        # Dislocated by race and ethnicity
        hud = self.pop_disl_result
        hud_vals = hud["dislocated"].value_counts()

        hud_vals_false = 0
        hud_vals_true = 0

        # if there is "False" category, assign the value
        if False in hud_vals.keys().values:
            hud_vals_false = int(hud_vals[False])

        # if there is "True" category, assign the value
        if True in hud_vals.keys().values:
            hud_vals_true = int(hud_vals[True])

        hua_disl = [hud_vals_false, hud_vals_true]

        pd_disl = [int(hud["numprec"].where(hud["dislocated"] == 0).sum()),
                   int(hud["numprec"].where(hud["dislocated"] == 1).sum())]

        hua_tot = sum(hua_disl)
        pop_tot = sum(pd_disl)

        no_hua_tot = {"households": None, "percent_of_households": None}
        hua_disl_tot = {}
        hua_disl_tot["dislocated"] = no_hua_tot
        hua_disl_tot["not_dislocated"] = no_hua_tot
        hua_disl_tot["total"] = no_hua_tot
        if hua_tot:
            hua_disl_tot["dislocated"] = {"households": hua_disl[1],
                                          "percent_of_households": 100 * (hua_disl[1]/hua_tot)}
            hua_disl_tot["not_dislocated"] = {"households": hua_tot - hua_disl[1],
                                              "percent_of_households": 100 * ((hua_tot - hua_disl[1])/hua_tot)}
            hua_disl_tot["total"] = {"households": hua_tot, "percent_of_households": 100}

        no_pop_tot = {"population": None, "percent_of_population": None}
        pop_disl_tot = {}
        pop_disl_tot["dislocated"] = no_pop_tot
        pop_disl_tot["not_dislocated"] = no_pop_tot
        pop_disl_tot["total"] = no_pop_tot
        if pop_tot:
            pop_disl_tot["dislocated"] = {"population": pd_disl[1],
                                          "percent_of_population": 100 * (pd_disl[1]/pop_tot)}
            pop_disl_tot["not_dislocated"] = {"population": pop_tot - pd_disl[1],
                                              "percent_of_population": 100 * ((pop_tot - pd_disl[1])/pop_tot)}
            pop_disl_tot["total"] = {"population": pop_tot, "percent_of_population": 100}

        pd_total_json = {"household_dislocation_in_total": hua_disl_tot,
                         "population_dislocation_in_total": pop_disl_tot}
        # print(pd_total_json)

        if filename_json:
            with open(filename_json, "w") as outfile:
                json.dump(pd_total_json, outfile)
        return json.dumps(pd_total_json)



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import numpy as np
import pandas as pd

from pyincore import BaseAnalysis
from pyincore.analyses.housingrecovery.housingrecoveryutil import HousingRecoveryUtil


class HousingRecovery(BaseAnalysis):
    """ The analysis predicts building values and value changes over time following a disaster event.
    The model is calibrated with respect to demographics, parcel data, and building value trajectories
    following Hurricane Ike (2008) in Galveston, Texas. The model predicts building value at the parcel
    level for 8 years of observation. The models rely on Census (Decennial or American Community Survey, ACS)
    and parcel data immediately prior to the disaster event (year -1) as inputs for prediction.

    The Galveston, TX example makes use of 2010 Decennial Census and Galveston County Appraisal District (GCAD)
    tax assessor data and outputs from other analysis (i.e., Building Damage, Housing Unit Allocation,
    Population Dislocation) .

    The CSV outputs of the building values for the 6 years following the disaster event (with year 0 being
    the impact year).

    Contributors
        | Science: Wayne Day, Sarah Hamideh
        | Implementation: Michal Ondrejcek, Santiago Núñez-Corrales, and NCSA IN-CORE Dev Team

    Related publications
        Hamideh, S., Peacock, W. G., & Van Zandt, S. (2018). Housing recovery after disasters: Primary versus
        seasonal/vacation housing markets in coastal communities. Natural Hazards Review.

    Args:
        incore_client (IncoreClient): Service authentication.

    """
    def __init__(self, incore_client):
        super(HousingRecovery, self).__init__(incore_client)

    def get_spec(self):
        return {
            "name": "housing-recovery",
            "description": "Housing Recovery Analysis",
            "input_parameters": [
                {
                    "id": "base_year",
                    "required": False,
                    "description": "Base year is used to calculate improvement age. It needs to be set to the tax "
                                   "assessment year representing pre-disaster building values. For example for GCAD "
                                   "data which represents improvement valuation before Hurricane Ike impacts."
                                   "Deafult 2008",
                    "type": int
                },
                {
                   "id": "result_name",
                   "required": True,
                   "description": "Result CSV dataset name",
                   "type": str
                }
            ],
            "input_datasets": [
                {
                   "id": "population_dislocation",
                   "required": True,
                   "description": "Population Dislocation aggregated to the block group level",
                   "type": ["incore:popDislocation"]
                },
                {
                   "id": "building_area",
                   "required": True,
                   "description": "Building square footage and damage. Damage is the actual building value loss "
                                  "in percentage terms observed through the County Appraisal District (GCAD) data",
                   "type": ["incore:buildingInventoryArea"]
                },
                {
                   "id": "census_block_groups_data",
                   "required": True,
                   "description": "Census ACS data, 2010 5yr data for block groups available at IPUMS NHGIS "
                                  "website.",
                   "type": ["incore:censusBlockGroupsData"]
                },
                {
                   "id": "census_appraisal_data",
                   "required": True,
                   "description": "Census data, 2010 Decennial Census District (GCAD) Census data",
                   "type": ["incore:censusAppraisalData"]
                }
            ],
            "output_datasets": [
                {
                   "id": "result",
                   "description": "A csv file with the building values for the 6 years following the disaster"
                                  "event (year -1 denotes pre-impact conditions and 0 being the impact year). "
                                  "Index year values represent building values against a base, pre-impact value.",
                   "type": "incore:buildingValues"
                }
            ]
        }

    def run(self):
        """ Executes the Housing recovery analysis.

        Returns:
            bool: True if successful, False otherwise.

        """
        hru = HousingRecoveryUtil()

        # Get base year
        self.base_year = self.get_parameter("base_year")
        if self.base_year is None:
            self.base_year = hru.BASEYEAR
        # Get desired result name
        result_name = self.get_parameter("result_name")

        # Datasets
        pop_disl = self.get_input_dataset("population_dislocation").get_dataframe_from_csv(low_memory=False)
        addl_structure_info = self.get_input_dataset("building_area").get_dataframe_from_csv(low_memory=False)
        bg_mhhinc = self.get_input_dataset("census_block_groups_data").get_dataframe_from_csv(low_memory=False)

        # Census data
        vac_status = self.get_input_dataset("census_appraisal_data").get_dataframe_from_csv(low_memory=False)

        # Calculate the percent vacation or seasonal housing of all housing units within a census tract
        vac_status = self.get_vac_season_housing(vac_status)

        # Get ownership by filtering vacancy codes
        pop_disl["d_ownerocc"] = self.get_owneship(pop_disl)

        # Keep only parcels with 1 building
        building_level = pop_disl.loc[(pop_disl.bldgobs == 1)]
        single_family = building_level.loc[(building_level.d_sf == 1)]

        # Read in and clean block group level median household income
        bg_mhhinc["mhhinck"] = bg_mhhinc["mhhinc"] / 1000

        # Read in and clean additional building information, NOTE add to building inventory
        # Create structure id for merging data
        addl_structure_info["strctid"] = addl_structure_info["xref"].apply(lambda x: "XREF"+x)
        hse_recov = self.merge_add_inv(single_family, addl_structure_info)

        # Merge with seasonal/vacation housing Census ACS data
        hse_recov = self.merge_seasonal_data(hse_recov, vac_status)

        # Merge with BG median HH income
        hse_recov = self.merge_block_data(hse_recov, bg_mhhinc)

        # Generate minority variable
        hse_recov["pminoritybg"] = hse_recov["phispbg"] + hse_recov["pblackbg"]

        # Estimate value_loss for each parcel based on parameters from Bai, Hueste, & Gardoni (2009)
        hse_recov = self.value_loss(hse_recov)

        # Primary Housing Market (PHM)
        hse_rec_phm = self.assemble_phm_coefs(hru, hse_recov)
        # Seasonal/Vacation housing market (SVHM)
        hse_rec_svhm = self.assemble_svhm_coefs(hru, hse_recov)

        d_vac_np = np.tile(hse_recov["d_vacationct"], (8, 1)).T
        # Vacation condition for all years

        hse_rec_fin = np.empty(hse_recov.shape)
        hse_rec_fin[:] = np.NaN
        hse_rec_fin = np.where(d_vac_np == 0, np.exp(hse_rec_phm), np.NaN)
        hse_rec_fin = np.where(d_vac_np == 1, np.exp(hse_rec_svhm), hse_rec_fin)

        # Index building values against a pre-impact (-1) value
        hse_rec_fin_index = hse_rec_fin / hse_rec_fin[:, 0:1]

        # Name columns, use list of damage years (range from -1 to n).
        bval_yr = ["bval_year_{0}".format(i) for i in hru.DMG_YEARS]
        index_yr = ["index_{0}".format(i) for i in hru.DMG_YEARS]

        hse_rec_fin1 = pd.DataFrame(hse_rec_fin, columns=bval_yr)
        hse_rec_fin2 = pd.DataFrame(hse_rec_fin_index, columns=index_yr)
        hse_recov = pd.concat([hse_recov, hse_rec_fin1, hse_rec_fin2], axis=1)

        columns_to_save = ["guid", "d_vacationct", "mhhinck", "pminoritybg", "dmg", "value_loss"] + bval_yr + index_yr
        self.set_result_csv_data("result", hse_recov[columns_to_save], result_name, "dataframe")

        return True

    def get_owneship(self, popd):
        """ Filter ownership based on the vacancy codes
        Assumption:
        Where ownershp is "missing", let vacancy codes 0/3/4 be considered owner-occupied,
        and 1/2/5/6/7 be considered renter-occupied.
        It is uncertain whether vacancy codes 3,4,5,6,7 will become owner- or renter-occupied or primarily
        one or the other.
    .
        Args:
            popd (pd.DataFrame): Population dislocation results with ownership information.

        Returns:
            pd.DataFrame: Ownership data.

        """
        # Create ownership dummy variable from popd.ownership
        own = np.where(popd["ownershp"].isnull() & popd["vacancy"] == 7, 0, np.nan)
        own = np.where(popd["ownershp"].isnull() & popd["vacancy"] == 6, 0, own)
        own = np.where(popd["ownershp"].isnull() & popd["vacancy"] == 5, 0, own)
        own = np.where(popd["ownershp"].isnull() & popd["vacancy"] == 2, 0, own)
        own = np.where(popd["ownershp"].isnull() & popd["vacancy"] == 1, 0, own)
        own = np.where(popd["ownershp"].isnull() & popd["vacancy"] == 4, 1, own)
        own = np.where(popd["ownershp"].isnull() & popd["vacancy"] == 3, 1, own)
        own = np.where(popd["ownershp"].isnull() & popd["vacancy"] == 0, 1, own)
        own = np.where(popd["ownershp"] == 2, 0, own)
        own = np.where(popd["ownershp"] == 1, 1, own)

        return own

    def get_vac_season_housing(self, vac_status):
        """ Calculate the percent vacation or seasonal housing of all housing units within a census tract and
        add dummy variable for census tract as a seasonal/vacation housing submarket.
.
        Args:
            vac_status (obj): Seasonal/vacation housing Census ACS data from json reader.

        Returns:
            pd.DataFrame: Seasonal/vacation housing data.

        """
        vac_status["B25004_006E"] = vac_status["B25004_006E"].astype(int)
        vac_status["B25004_006M"] = vac_status["B25004_006M"].astype(int)
        vac_status["B25002_001E"] = vac_status["B25002_001E"].astype(int)
        vac_status["B25002_001M"] = vac_status["B25002_001M"].astype(int)

        # Calculate the percent vacation or seasonal housing of all housing units within a census tract
        vac_status["pvacationct_moe"] = vac_status["B25004_006E"] / vac_status["B25002_001E"]
        vac_status["pvacationct"] = 100 * vac_status["pvacationct_moe"]
        vac_status["pvacationct_moe"] = vac_status["pvacationct_moe"] ** 2 * vac_status["B25002_001M"] ** 2
        vac_status["pvacationct_moe"] = vac_status["B25004_006M"] ** 2 - vac_status["pvacationct_moe"]
        vac_status["pvacationct_moe"] = 100 * (1 / vac_status["B25002_001E"]) * vac_status["pvacationct_moe"] ** 0.5

        # dummy variable for census tract as a seasonal/vacation housing submarket
        vac_status["d_vacationct"] = np.where(vac_status["pvacationct"] >= 50, 1, 0)

        vac_status.sort_values(by="pvacationct", inplace=True, ascending=False)

        return vac_status

    def merge_add_inv(self, hse_rec, addl_struct):
        """Merge study area and additional structure information.
.
        Args:
            hse_rec (pd.DataFrame):  Area inventory.
            addl_struct (pd.DataFrame):  Additional infrastructure inventory.

        Returns:
            pd.DataFrame: Final merge of two inventories.

        """
        hse_rec_merged = pd.merge(hse_rec, addl_struct, on="strctid", how="inner")
        return hse_rec_merged

    def merge_seasonal_data(self, hse_rec, vac_status):
        """ Merge study area and with seasonal/vacation housing Census ACS data.
.
        Args:
            hse_rec (pd.DataFrame): Area inventory.
            vac_status (pd.DataFrame): Seasonal/vacation housing Census ACS data.

        Returns:
            pd.DataFrame: Final merge of two inventories.

        """
        hse_rec["tractid"] = hse_rec["tractid"].astype(str)
        # Add county and state to trac to match hse_rec tracid (Galveston - 723900 to 48167723900)
        vac_status["tractid"] = \
            vac_status["state"].astype(str) + vac_status["county"].astype(str) + vac_status["tract"].astype(str)

        hse_rec_merged = pd.merge(hse_rec, vac_status, left_on="tractid", right_on="tractid", how='inner')
        return hse_rec_merged

    def merge_block_data(self, hse_rec, bg_mhhinc):
        """ Merge block group level median household income.
.
        Args:
            hse_rec (pd.DataFrame):  Area inventory.
            bg_mhhinc (pd.DataFrame):  Block data.

        Returns:
            pd.DataFrame: Final merge of two inventories.

        """
        hse_rec_merged = pd.merge(hse_rec, bg_mhhinc, left_on="bgidstr", right_on="bgidstr", how="inner")
        return hse_rec_merged

    def value_loss(self, hse_rec):
        """ Estimate value_loss for each parcel based on parameters from Bai, Hueste, & Gardoni (2009).
    .
        Args:
            hse_rec (pd.DataFrame):  Area inventory.

        Returns:
            pd.DataFrame: Inventory with value losses.

        """
        hse_rec["value_loss"] = 100 * (hse_rec["DS_0"] * hse_rec["rploss_0"] +
                                       hse_rec["DS_1"] * hse_rec["rploss_1"] +
                                       hse_rec["DS_2"] * hse_rec["rploss_2"] +
                                       hse_rec["DS_3"] * hse_rec["rploss_3"])
        return hse_rec

    def assemble_phm_coefs(self, hru, hse_rec):
        """ Assemble Primary Housing Market (PHM) data for full inventory and all damage-related years.
    .
        Args:
            hru (obj): Housing recovery utility.
            hse_rec (pd.DataFrame):  Area inventory including losses.

        Returns:
            np.array: Final coefficients for all damage years.

        """
        dmg_years = np.array(hru.DMG_YEARS)
        dmg_years_size = len(hru.DMG_YEARS)

        coef_fin = np.empty((hse_rec.shape[0], dmg_years_size))
        coef_fin[:] = hru.B_PHM_intercept + np.fromiter(hru.B_PHM_year.values(), dtype=float)

        # Adjust build year year with damage years
        yrbl_all = np.empty((hse_rec.shape[0], dmg_years_size))
        yrbl_all[:] = self.base_year + dmg_years + 1
        yrbuilt = hru.B_PHM_age * (yrbl_all - hse_rec["effyrbuilt"].to_numpy()[:, np.newaxis])

        # Square meters, use vector (1x8) with B_PHM_sqm
        sqmeter = np.full((1, dmg_years_size), hru.B_PHM_sqm) * hse_rec["sqmeter"].to_numpy()[:, np.newaxis]

        if "dmg" in hse_rec.columns:
            dmg_loss = np.fromiter(hru.B_PHM_dmg_year.values(), dtype=float) * \
                       hse_rec["dmg"].to_numpy()[:, np.newaxis]
        else:
            dmg_loss = np.fromiter(hru.B_PHM_dmg_year.values(), dtype=float) * \
                       hse_rec["value_loss"].to_numpy()[:, np.newaxis]
        d_owner = \
            np.fromiter(hru.B_PHM_own_year.values(), dtype=float) * hse_rec["d_ownerocc"].to_numpy()[:, np.newaxis]
        mhhinck = \
            np.fromiter(hru.B_PHM_inc_year.values(), dtype=float) * hse_rec["mhhinck"].to_numpy()[:, np.newaxis]
        pminrbg = \
            np.fromiter(hru.B_PHM_min_year.values(), dtype=float) * hse_rec["pminoritybg"].to_numpy()[:, np.newaxis]

        return coef_fin + yrbuilt + sqmeter + d_owner + dmg_loss + mhhinck + pminrbg

    def assemble_svhm_coefs(self, hru, hse_rec):
        """ Assemble Seasonal/Vacation housing market (SVHM) data for full inventory and all damage-related years.
    .
        Args:
            hru (obj): Housing recovery utility.
            hse_rec (pd.DataFrame):  Area inventory including losses.

        Returns:
            np.array: Final coefficients for all damage years.

        """
        dmg_years = np.array(hru.DMG_YEARS)
        dmg_years_size = len(hru.DMG_YEARS)

        coef_fin = np.empty((hse_rec.shape[0], dmg_years_size))
        coef_fin[:] = hru.B_SVHM_intercept + np.fromiter(hru.B_SVHM_year.values(), dtype=float)

        # Adjust build year year with damage years
        yrbl_all = np.empty((hse_rec.shape[0], dmg_years_size))
        yrbl_all[:] = self.base_year + dmg_years + 1
        yrbuilt = hru.B_SVHM_age * (yrbl_all - hse_rec["effyrbuilt"].to_numpy()[:, np.newaxis])

        # Square meters, use vector (1x8) with B_PHM_sqm
        sqmeter = np.full((1, dmg_years_size), hru.B_SVHM_sqm) * hse_rec["sqmeter"].to_numpy()[:, np.newaxis]

        if "dmg" in hse_rec.columns:
            dmg_loss = np.fromiter(hru.B_SVHM_dmg_year.values(), dtype=float) * \
                       hse_rec["dmg"].to_numpy()[:, np.newaxis]
        else:
            dmg_loss = np.fromiter(hru.B_SVHM_dmg_year.values(), dtype=float) * \
                       hse_rec["value_loss"].to_numpy()[:, np.newaxis]
        d_owner = \
            np.fromiter(hru.B_SVHM_own_year.values(), dtype=float) * hse_rec["d_ownerocc"].to_numpy()[:, np.newaxis]
        mhhinck = \
            np.fromiter(hru.B_SVHM_inc_year.values(), dtype=float) * hse_rec["mhhinck"].to_numpy()[:, np.newaxis]

        return coef_fin + yrbuilt + sqmeter + dmg_loss + d_owner + mhhinck



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import collections
import concurrent.futures
import numpy as np

from typing import List
from pyincore import BaseAnalysis, AnalysisUtil


class MonteCarloFailureProbability(BaseAnalysis):
    """
    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        super(MonteCarloFailureProbability, self).__init__(incore_client)

    def get_spec(self):
        """Get specifications of the monte carlo failure probability analysis.

        Returns:
            obj: A JSON object of specifications of the monte carlo failure probability analysis.

        """
        return {
            'name': 'monte-carlo-failure-probability',
            'description': 'calculate the probability of failure in monte-carlo simulation',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'basename of the result datasets. This analysis will create two outputs: failure '
                                   'proability and failure state with the basename in the filename. '
                                   'For example: "result_name = joplin_mcs_building" will result in '
                                   '"joplin_mcs_building_failure_state.csv" and '
                                   '"joplin_mcs_building_failure_probability.csv"',
                    'type': str
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                },
                {
                    'id': 'num_samples',
                    'required': True,
                    'description': 'Number of MC samples',
                    'type': int
                },
                {
                    'id': 'damage_interval_keys',
                    'required': True,
                    'description': 'Column name of the damage interval',
                    'type': List[str]
                },
                {
                    'id': 'failure_state_keys',
                    'required': True,
                    'description': 'Column name of the damage interval that considered as damaged',
                    'type': List[str]
                },
                {
                    'id': 'seed',
                    'required': False,
                    'description': 'Initial seed for the probabilistic model',
                    'type': int
                },
            ],
            'input_datasets': [
                {
                    'id': 'damage',
                    'required': True,
                    'description': 'damage result that has damage intervals in it',
                    'type': ['ergo:buildingDamageVer4',
                             'ergo:buildingDamageVer5',
                             'ergo:buildingDamageVer6',
                             'ergo:nsBuildingInventoryDamage',
                             'ergo:nsBuildingInventoryDamageVer2',
                             'ergo:nsBuildingInventoryDamageVer3',
                             'ergo:bridgeDamage',
                             'ergo:bridgeDamageVer2',
                             'ergo:bridgeDamageVer3',
                             'ergo:waterFacilityDamageVer4',
                             'ergo:waterFacilityDamageVer5',
                             'ergo:waterFacilityDamageVer6',
                             'ergo:roadDamage',
                             'ergo:roadDamageVer2',
                             'ergo:roadDamageVer3',
                             'incore:epfDamage',
                             'incore:epfDamageVer2',
                             'incore:epfDamageVer3',
                             'incore:pipelineDamage',
                             'incore:pipelineDamageVer2',
                             'incore:pipelineDamageVer3']
                },

            ],
            'output_datasets': [
                {
                    'id': 'failure_probability',
                    'description': 'CSV file of failure probability',
                    'type': 'incore:failureProbability'
                },
                {
                    'id': 'sample_failure_state',
                    'description': 'CSV file of failure state for each sample',
                    'type': 'incore:sampleFailureState'
                },
                {
                    'id': 'sample_damage_states',
                    'description': 'CSV file of simulated damage states for each sample',
                    'type': 'incore:sampleDamageState'
                }
            ]
        }

    def run(self):
        """Executes mc failure probability analysis."""

        # read in file and parameters
        damage = self.get_input_dataset("damage").get_csv_reader()
        damage_result = AnalysisUtil.get_csv_table_rows(damage, ignore_first_row=False)

        # setting number of cpus to use
        user_defined_cpu = 1
        if not self.get_parameter("num_cpu") is None and self.get_parameter(
                "num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        num_workers = AnalysisUtil.determine_parallelism_locally(self,
                                                                 len(
                                                                     damage_result),
                                                                 user_defined_cpu)

        avg_bulk_input_size = int(len(damage_result) / num_workers)
        inventory_args = []
        count = 0
        inventory_list = damage_result

        seed = self.get_parameter("seed")
        seed_list = []
        if seed is not None:
            while count < len(inventory_list):
                inventory_args.append(
                    inventory_list[count:count + avg_bulk_input_size])
                seed_list.append([seed + i for i in range(count - 1, count + avg_bulk_input_size - 1)])
                count += avg_bulk_input_size
        else:
            while count < len(inventory_list):
                inventory_args.append(
                    inventory_list[count:count + avg_bulk_input_size])
                seed_list.append([None for i in range(count - 1, count + avg_bulk_input_size - 1)])
                count += avg_bulk_input_size

        fs_results, fp_results, samples_results = self.monte_carlo_failure_probability_concurrent_future(
            self.monte_carlo_failure_probability_bulk_input, num_workers,
            inventory_args, seed_list)
        self.set_result_csv_data("sample_failure_state",
                                 fs_results, name=self.get_parameter("result_name") + "_failure_state")
        self.set_result_csv_data("failure_probability",
                                 fp_results, name=self.get_parameter("result_name") + "_failure_probability")
        self.set_result_csv_data("sample_damage_states",
                                 samples_results, name=self.get_parameter("result_name") + "_sample_damage_states")
        return True

    def monte_carlo_failure_probability_concurrent_future(self, function_name,
                                                          parallelism, *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            parallelism (int): Number of workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            list: A list of dictionary with id/guid and failure state for N samples.
            list: A list dictionary with failure probability and other data/metadata.

        """
        fs_output = []
        fp_output = []
        samples_output = []
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=parallelism) as executor:
            for fs_ret, fp_ret, samples_ret in executor.map(function_name, *args):
                fs_output.extend(fs_ret)
                fp_output.extend(fp_ret)
                samples_output.extend(samples_ret)

        return fs_output, fp_output, samples_output

    def monte_carlo_failure_probability_bulk_input(self, damage, seed_list):
        """Run analysis for monte carlo failure probability calculation

        Args:
            damage (obj): An output of building/bridge/waterfacility/epn damage that has damage interval.
            seed_list (list): Random number generator seed per building for reproducibility.

        Returns:
            fs_results (list): A list of dictionary with id/guid and failure state for N samples
            fp_results (list): A list dictionary with failure probability and other data/metadata.

        """
        damage_interval_keys = self.get_parameter("damage_interval_keys")
        failure_state_keys = self.get_parameter("failure_state_keys")
        num_samples = self.get_parameter("num_samples")

        fs_result = []
        fp_result = []
        samples_output = []

        i = 0
        for dmg in damage:
            fs, fp, samples_result = self.monte_carlo_failure_probability(dmg, damage_interval_keys, failure_state_keys,
                                                                          num_samples, seed_list[i])
            fs_result.append(fs)
            fp_result.append(fp)
            samples_output.append(samples_result)
            i += 1

        return fs_result, fp_result, samples_output

    def monte_carlo_failure_probability(self, dmg, damage_interval_keys,
                                        failure_state_keys, num_samples, seed):
        """Calculates building damage results for a single building.

        Args:
            dmg (obj): Damage analysis output for a single entry.
            damage_interval_keys (list): A list of the name of the damage intervals.
            failure_state_keys (list): A list of the name of the damage state that is considered as failed.
            num_samples (int): Number of samples for mc simulation.
            seed (int): Random number generator seed for reproducibility.

        Returns:
            dict: A dictionary with id/guid and failure state for N samples
            dict: A dictionary with failure probability and other data/metadata.
            dict: A dictionary with id/guid and damage states for N samples

        """
        # failure state
        fs_result = collections.OrderedDict()

        # sample damage states
        samples_result = collections.OrderedDict()

        # copying guid/id column to the sample damage failure table
        if 'guid' in dmg.keys():
            fs_result['guid'] = dmg['guid']
            samples_result['guid'] = dmg['guid']

        elif 'id' in dmg.keys():
            fs_result['id'] = dmg['id']
            samples_result['id'] = dmg['id']
        else:
            fs_result['id'] = 'NA'
            samples_result['id'] = 'NA'

        # failure probability
        fp_result = collections.OrderedDict()
        fp_result['guid'] = dmg['guid']

        ds_sample = self.sample_damage_interval(dmg, damage_interval_keys,
                                                num_samples, seed)
        func, fp = self.calc_probability_failure_value(ds_sample, failure_state_keys)

        fs_result['failure'] = ",".join(func.values())
        fp_result['failure_probability'] = fp
        samples_result['sample_damage_states'] = ','.join(ds_sample.values())

        return fs_result, fp_result, samples_result

    def sample_damage_interval(self, dmg, damage_interval_keys, num_samples, seed):
        """
        Dylan Sanderson code to calculate the Monte Carlo simulations of damage state.

        Args:
            dmg (dict): Damage results that contains dmg interval values.
            damage_interval_keys (list): Keys of the damage states.
            num_samples (int): Number of simulation.
            seed (int): Random number generator seed for reproducibility.

        Returns:
            dict: A dictionary of damage states.

        """
        ds = {}
        random_generator = np.random.RandomState(seed)
        for i in range(num_samples):
            # each sample should have a unique seed
            rnd_num = random_generator.uniform(0, 1)
            prob_val = 0
            flag = True
            for ds_name in damage_interval_keys:

                if rnd_num < prob_val + AnalysisUtil.float_to_decimal(dmg[ds_name]):
                    ds['sample_{}'.format(i)] = ds_name
                    flag = False
                    break
                else:
                    prob_val += AnalysisUtil.float_to_decimal(dmg[ds_name])
            if flag:
                print("cannot determine MC damage state!")
                break

        return ds

    def calc_probability_failure_value(self, ds_sample, failure_state_keys):
        """
        Lisa Wang's approach to calculate a single value of failure probability.

        Args:
            ds_sample (dict): A dictionary of damage states.
            failure_state_keys (list): Damage state keys that considered as failure.

        Returns:
            float: Failure state on each sample 0 (failed), 1 (not failed).
            float: Failure probability (0 - 1).

        """
        count = 0
        func = {}
        for sample, state in ds_sample.items():
            if state in failure_state_keys:
                func[sample] = "0"
                count += 1
            else:
                func[sample] = "1"
        if len(ds_sample):
            return func, count / len(ds_sample)
        else:
            return func, np.nan



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

from logging import config as logging_config

import logging
import os
import shutil

PACKAGE_VERSION = "1.15.0"

INCORE_API_PROD_URL = "https://incore.ncsa.illinois.edu"
INCORE_API_DEV_URL = "https://incore-dev.ncsa.illinois.edu"

KEYCLOAK_AUTH_PATH = "/auth/realms/In-core/protocol/openid-connect/token"
KEYCLOAK_USERINFO_PATH = "/auth/realms/In-core/protocol/openid-connect/userinfo"
CLIENT_ID = "react-auth"
INCORE_LDAP_TEST_USER_INFO = "{\"preferred_username\": \"incrtest\"}"

PYINCORE_PACKAGE_HOME = os.path.dirname(__file__)
PYINCORE_ROOT_FOLDER = os.path.dirname(os.path.dirname(__file__))
USER_HOME = os.path.expanduser('~')
USER_CACHE_DIR = ".incore"
PYINCORE_USER_CACHE = os.path.join(USER_HOME, USER_CACHE_DIR)

DATA_CACHE_FOLDER_NAME = "cache_data"
DATA_CACHE_HASH_NAMES_SERVICE_JSON = "service.json"
PYINCORE_USER_DATA_CACHE = os.path.join(PYINCORE_USER_CACHE, DATA_CACHE_FOLDER_NAME)
PYINCORE_SERVICE_JSON = os.path.join(PYINCORE_USER_CACHE, DATA_CACHE_HASH_NAMES_SERVICE_JSON)

LOGGING_CONFIG = os.path.abspath(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'logging.ini'))
logging_config.fileConfig(LOGGING_CONFIG)
LOGGER = logging.getLogger('pyincore')

TEST_DATA_DIR = os.path.join(PYINCORE_ROOT_FOLDER, "tests/data")

MAX_LOGIN_ATTEMPTS = 3

IPOPT_PATH = shutil.which("ipopt")
GLPK_PATH = shutil.which("glpsol")
SCIP_PATH = shutil.which("scip")

DAMAGE_PRECISION = 10



"""
These function compute the change of demand values over time based on population dislocation models for Seaside and
dislocation time model for Lumberton.
"""

import math
import os
import pickle

import numpy as np
import pandas as pd


class DislocationUtil:

    @staticmethod
    def create_dynamic_param(params, pop_dislocation, dt_params, T=1, N=None):
        """
        This function computes the change of demand values over time based on population dislocation models.

        Parameters
        ----------
        params : dict
            Parameters that are needed to run the INDP optimization.
        T : int, optional
            Number of time steps to optimize over. T=1 shows an iINDP analysis, and T>1 shows a td-INDP. The default is
            1.
        N : :class:`~infrastructure.InfrastructureNetwork`
            The object containing the network data.

        Returns
        -------
         dynamic_params : dict
             Dictionary of dynamic demand value for nodes
        """
        dynamic_param_dict = params['DYNAMIC_PARAMS']
        return_type = dynamic_param_dict['RETURN']
        dp_dict_col = ['time', 'node', 'current pop', 'total pop']
        net_names = {'WATER': 1, 'GAS': 2, 'POWER': 3, 'TELECOME': 4}
        dynamic_params = {}

        output_file = dynamic_param_dict['TESTBED'] + '_pop_dislocation_demands_' + str(params['MAGNITUDE']) + 'yr.pkl'
        if os.path.exists(output_file):
            print("\nReading from file...")
            with open(output_file, 'rb') as f:
                dynamic_params = pickle.load(f)
            return dynamic_params
        for net in dynamic_param_dict['MAPPING'].keys():
            nn = net_names[net]
            mapping_data = dynamic_param_dict['MAPPING'][net]
            dynamic_params[nn] = pd.DataFrame(columns=dp_dict_col)
            if net == 'POWER':
                for n, d in N.G.nodes(data=True):
                    if n[1] != nn:
                        continue
                    guid = d['data']['inf_data'].guid
                    # Find building in the service area of the node/substation
                    try:
                        serv_area = mapping_data[mapping_data['substations_guid'] == guid]
                    except KeyError:
                        serv_area = mapping_data[mapping_data['node_guid'] == guid]
                    # compute dynamic_params
                    num_dilocated = {t: 0 for t in range(T + 1)}
                    total_pop_node = 0
                    for _, bldg in serv_area.iterrows():
                        try:
                            pop_bldg_dict = pop_dislocation[pop_dislocation['guid'] == bldg['buildings_guid']]
                        except KeyError:
                            pop_bldg_dict = pop_dislocation[pop_dislocation['guid'] == bldg['bldg_guid']]
                        for _, hh in pop_bldg_dict.iterrows():
                            total_pop_node += hh['numprec'] if ~np.isnan(hh['numprec']) else 0
                            if hh['dislocated']:
                                # ..todo Lumebrton dislocation time model. Replace with that of Seaside when available
                                return_time = DislocationUtil.disloc_time_mode(hh, dt_params)
                                for t in range(return_time):
                                    if t <= T and return_type == 'step_function':
                                        num_dilocated[t] += hh['numprec'] if ~np.isnan(hh['numprec']) else 0
                                    elif t <= T and return_type == 'linear':
                                        pass  # ..todo Add linear return here
                    for t in range(T + 1):
                        values = [t, n[0], total_pop_node - num_dilocated[t], total_pop_node]
                        dynamic_params[n[1]] = dynamic_params[n[1]].append(dict(zip(dp_dict_col, values)),
                                                                           ignore_index=True)
            elif net == 'WATER':
                node_pop = {}
                for u, v, a in N.G.edges(data=True):
                    if u[1] != nn or v[1] != nn:
                        continue
                    guid = a['data']['inf_data'].guid
                    # Find building in the service area of the pipe
                    serv_area = mapping_data[mapping_data['edge_guid'] == guid]
                    start_node = u[0]
                    if start_node not in node_pop.keys():
                        node_pop[start_node] = {'total_pop_node': 0,
                                                'num_dilocated': {t: 0 for t in range(T + 1)}}
                    end_node = v[0]
                    if end_node not in node_pop.keys():
                        node_pop[end_node] = {'total_pop_node': 0,
                                              'num_dilocated': {t: 0 for t in range(T + 1)}}
                    # compute dynamic_params
                    for _, bldg in serv_area.iterrows():
                        pop_bldg_dict = pop_dislocation[pop_dislocation['guid'] == bldg['bldg_guid']]
                        for _, hh in pop_bldg_dict.iterrows():
                            # half of the arc's demand is assigned to each node
                            # also, each arc is counted twice both as (u,v) and (v,u)
                            node_pop[start_node]['total_pop_node'] += hh['numprec'] / 4 \
                                if ~np.isnan(hh['numprec']) else 0
                            node_pop[end_node]['total_pop_node'] += hh['numprec'] / 4 \
                                if ~np.isnan(hh['numprec']) else 0
                            if hh['dislocated']:
                                # ..todo Lumebrton dislocation time model. Replace with that of Seaside when available
                                return_time = DislocationUtil.disloc_time_mode(hh, dt_params)
                                for t in range(return_time):
                                    if t <= T and return_type == 'step_function':
                                        node_pop[start_node]['num_dilocated'][t] += hh['numprec'] / 4 \
                                            if ~np.isnan(hh['numprec']) else 0
                                        node_pop[end_node]['num_dilocated'][t] += hh['numprec'] / 4 \
                                            if ~np.isnan(hh['numprec']) else 0
                                    elif t <= T and return_type == 'linear':
                                        pass  # ..todo Add linear return here
                for n, val in node_pop.items():
                    for t in range(T + 1):
                        values = [t, n, val['total_pop_node'] - val['num_dilocated'][t],
                                  val['total_pop_node']]
                        dynamic_params[nn] = dynamic_params[nn].append(dict(zip(dp_dict_col, values)),
                                                                       ignore_index=True)
        with open(output_file, 'wb') as f:
            pickle.dump(dynamic_params, f)
        return dynamic_params

    @staticmethod
    def disloc_time_mode(household_data, dt_params):
        race_white = 1 if household_data['race'] == 1 else 0
        race_black = 1 if household_data['race'] == 2 else 0
        hispan = household_data['hispan'] if ~np.isnan(household_data['hispan']) else 0
        # ..todo verify that the explanatory variable correspond to columns in dt_params
        # ..todo Replace random insurance assumption
        linear_term = household_data['DS_0'] * dt_params['DS0'] + household_data['DS_1'] * dt_params['DS1'] + \
            household_data['DS_2'] * dt_params['DS2'] + household_data['DS_3'] * dt_params[
                          'DS3'] + race_white * dt_params['white'] + race_black * dt_params['black'] + hispan * \
            dt_params['hispanic'] + np.random.choice([0, 1], p=[.15, .85]) * dt_params['insurance']
        # household_data['randincome']/1000*dt_params['income']+\#!!! income data
        disloc_time = np.exp(linear_term)
        return_time = math.ceil(disloc_time / 7)  # !!! assuming each time step is one week
        return return_time

    @staticmethod
    def dynamic_parameters(N, original_N, t, dynamic_params):
        for n, d in N.G.nodes(data=True):
            data = dynamic_params[d['data']['inf_data'].net_id]
            if d['data']['inf_data'].demand < 0:
                current_pop = data.loc[(data['node'] == n[0]) & (data['time'] == t), 'current pop'].iloc[0]
                total_pop = data.loc[(data['node'] == n[0]) & (data['time'] == t), 'total pop'].iloc[0]
                original_demand = original_N.G.nodes[n]['data']['inf_data'].demand
                d['data']['inf_data'].demand = original_demand * current_pop / total_pop



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import math
import csv
import numpy as np
import scipy as sp
import scipy.stats
import concurrent.futures

from pyincore.analyses.buildingportfolio.BuildingData import BuildingData
from pyincore import BaseAnalysis, Dataset


class BuildingPortfolioRecoveryAnalysis(BaseAnalysis):
    """The Building Portfolio Recovery analysis uses damage probabilities of structural components,
    nonstructural drift-sensitive components, and nonstructural acceleration-sensitive components
    to calculate building’s initial functionality loss.

    Args:
        incore_client (IncoreClient): Service authentication.

    """
    def __init__(self, incore_client):
        super(BuildingPortfolioRecoveryAnalysis, self).__init__(incore_client)

    def get_spec(self):
        return {
            'name': 'building-portfolio-damage-analysis',
            'description': 'Building Portfolio Damage Analysis (with uncertainty)',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': False,
                    'description': 'Result dataset name',
                    'type': str
                },
                {
                    'id': 'uncertainty',
                    'required': True,
                    'description': 'Use uncertainty',
                    'type': bool
                },
                {
                    'id': 'sample_size',
                    'required': False,
                    'description': 'No. of buildings to be considered from input buildings',
                    'type': int
                },
                {
                    'id': 'random_sample_size',
                    'required': True,
                    'description': 'Number of iterations for Monte Carlo Simulation',
                    'type': int
                },
                {
                    'id': 'no_of_weeks',
                    'required': True,
                    'description': 'Number of weeks to run the recovery model',
                    'type': int
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                }
            ],
            'input_datasets': [
                {
                    'id': 'building_data',
                    'required': True,
                    'description': 'Building Data',
                    'type': ['incore:portfolioBuildingInventory'],
                },
                {
                    'id': 'occupancy_mapping',
                    'required': True,
                    'description': 'Occupancy code mapping',
                    'type': ['incore:portfolioOccupancyMapping'],
                },
                {
                    'id': 'building_damage',
                    'required': True,
                    'description': 'Building Damage Results',
                    'type': ['incore:portfolioBuildingDamage'],
                },
                {
                    'id': 'dmg_ratios',
                    'required': True,
                    'description': 'Percentage of mean repair by occupancy / building type',
                    'type': ['incore:portfolioDamageRatios'],
                },
                {
                    'id': 'utility',
                    'required': True,
                    'description': 'Full utility availability at each utility service area - joint area of power'
                                   'and water (row), at each week (column)"',
                    'type': ['incore:portfolioUtilityAvailability'],
                },
                {
                    'id': 'utility_partial',
                    'required': True,
                    'description': 'Partial utility availability at each utility service area',
                    'type': ['incore:portfolioUtilityAvailability'],
                },
                {
                    'id': 'coefFL',
                    'required': True,
                    'description': 'Correlation coefficient of initial functionality states',
                    'type': ['incore:portfolioCoefficients'],
                },
            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'buildingPortfolio',
                    'description': 'Building portfolio recovery result.',
                    'type': 'incore:portfolioRecovery'
                }
            ]
        }

    def run(self):

        uncertainty = self.get_parameter("uncertainty")
        utility_initial = self.get_input_dataset("utility").get_dataframe_from_csv()
        building_damage_results = self.get_input_dataset("building_damage").get_dataframe_from_csv()
        building_data = self.get_input_dataset("building_data").get_dataframe_from_csv()
        mean_repair = self.get_input_dataset("dmg_ratios").get_dataframe_from_csv()
        occupancy_mapping = self.get_input_dataset("occupancy_mapping").get_dataframe_from_csv()
        coeFL = self.get_input_dataset("coefFL").get_dataframe_from_csv()

        print('INFO: Data for Building Portfolio Recovery Analysis loaded successfully.')

        sample_size = self.get_parameter("sample_size")  # len(building  _damage_results)
        if sample_size is None:
            sample_size = len(building_damage_results)
        else:
            building_damage_results = building_damage_results.head(sample_size)
            coeFL = coeFL.iloc[0:sample_size, 0:sample_size]

        user_defined_cpu = self.get_parameter("num_cpu")
        permutation = np.random.permutation(len(building_data))
        permutation_subset = permutation[0:sample_size]
        sample_buildings = [BuildingData(building_data['Tract_ID'][i], building_data['X_Lon'][i],
                                         building_data['Y_Lat'][i], building_data['Structural'][i],
                                         building_data['Code_Level'][i], building_data['EPSANodeID'][i],
                                         building_data['PWSANodeID'][i], building_data['TEP_ID'][i],
                                         building_data['Build_ID_X'][i], building_data['EPSAID'][i],
                                         building_data['PWSAID'][i], building_data['Finance'][i],
                                         building_data['EP_PW_ID'][i], building_data['Occu_Code'][i]
                                         )
                            for i in permutation_subset]
        occupancy_map = {occupancy_mapping["Occu_ID"][i]: occupancy_mapping['Occupancy'][i] for i in
                         range(len(occupancy_mapping))}
        repair_mean = {mean_repair['Occupancy'][i]: [mean_repair['RC1'][i], mean_repair['RC2'][i],
                                                     mean_repair['RC3'][i], mean_repair['RC4'][i]]
                       for i in range(len(mean_repair))}

        building_damage = [[building_damage_results['Restricted Entry'][i],
                            building_damage_results['Restricted Use'][i],
                            building_damage_results['Reoccupancy'][i],
                            building_damage_results['Best Line Functionality'][i],
                            building_damage_results['Full Functionality'][i]]
                           for i in range(len(building_damage_results))]

        # START: Calculate waiting time statistics using Monte Carlo Simulations
        nsd = 5000  # TODO: Input?
        number_of_simulations = self.get_parameter("random_sample_size")

        output_base_name = self.get_parameter("result_name")
        if output_base_name is None:
            output_base_name = ""
        sample_delay = np.zeros(nsd)
        impeding_mean = np.zeros((5, 4))
        impeding_std = np.zeros((5, 4))

        for i in range(4):
            for j in range(5):
                for k in range(nsd):
                    sample_delay[k] = self.calculate_delay_time(i, j)
                impeding_mean[j, i] = np.mean(sample_delay)
                impeding_std[j, i] = np.std(sample_delay)

        # END: Calculate waiting time statistics using Monte Carlo Simulations

        # Recovery time step from week 1 to n
        time_steps = self.get_parameter("no_of_weeks")
        utility = np.ones((len(utility_initial), time_steps))

        # START initializing variables for uncertainty analysis
        target_mean_recovery = np.zeros(time_steps)
        utility2 = np.zeros((len(utility_initial), time_steps))
        # END code for uncertainty analysis

        # Utility matrix. Represents the recovery for each of the utility areas at a given number of weeks after the event
        # The utility matrix uses utility Initial input file for the first 22 weeks and ones for the rest of the time.

        if uncertainty:
            for i in range(time_steps):
                utility[:, i] = [1 - j for j in utility_initial[str(i)]]
        else:
            for i in range(time_steps):
                utility[:, i] = [j for j in utility_initial[str(i)]]

        # with concurrent.futures.ProcessPoolExecutor(max_workers=sample_size) as executor:
        #     for response in executor.map(self.calculate_transition_probability_matrix(sample_size, time_steps, sample_buildings,
        #                                                                          repair_mean, occupancy_map, uncertainty,
        #                                                                          impeding_mean, impeding_std,
        #                                                                          building_damage, utility, utility2)):
        response = self.calculate_transition_probability_matrix(time_steps, sample_buildings, repair_mean,
                                                                occupancy_map, uncertainty, impeding_mean, impeding_std,
                                                                building_damage, utility, utility2)
        temporary_correlation1 = response["temporary_correlation1"]
        temporary_correlation2 = response["temporary_correlation2"]
        mean_over_time = response["mean_over_time"]
        variance_over_time = response["variance_over_time"]

        # Functionality Recovery (Best Line Functionality + Full functionality) for each building on the sample
        recovery_fp = response["recovery_fp"]

        # Mean recovery trajectory at portfolio level
        mean_recovery = response["mean_recovery"]

        # Trajectory for the Restricted entry, Restricted Use, Reoccupancy, Best Line Functionality, Full Functionality
        mean_recovery = mean_recovery / sample_size

        # Trajectory for Best Line Functionality and Full Functionality
        mean_recovery_output = sum(recovery_fp) / sample_size

        with open(output_base_name + '_building-recovery.csv', 'w+', newline='') as output_file:
            spam_writer = csv.writer(output_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            spam_writer.writerow(['Building_ID', 'Building_Lon', 'Building_Lat']
                                 + list(range(1, time_steps + 1)))
            for i in range(sample_size):
                spam_writer.writerow([building_data['Build_ID_X'][i], building_data['X_Lon'][i],
                                      building_data['Y_Lat'][i]] + list(recovery_fp[i]))

        if uncertainty:
            # START: Additional Code for uncertainty analysis
            mean_u = np.zeros(sample_size)
            covar = np.matrix(coeFL)  # np.zeros([sample_size, sample_size])
            random_distribution = np.random.multivariate_normal(mean_u, covar, 10000)
            random_samples = sp.stats.norm.cdf(random_distribution)

            sample_total = self.calculate_sample_total(number_of_simulations, sample_size,
                                                       building_damage, random_samples)

            for k in range(sample_size):
                for t in range(time_steps):
                    if variance_over_time[t][k] <= 0:
                        variance_over_time[t][k] = 0

            # Start calculating standard deviation of the mean recovery trajectory
            total_standard_deviation = self.calculate_std_of_mean_bulk_input(range(time_steps), sample_size,
                                                                             number_of_simulations, variance_over_time,
                                                                             mean_over_time, temporary_correlation1,
                                                                             temporary_correlation2, sample_total)

            # Calculate distribution of Portfolio Recovery Time (PRT) assume normal distribution
            x_range = np.arange(0.0, 1.001, 0.001)
            pdf_full = np.zeros((time_steps, len(x_range)))
            irt = np.zeros(time_steps)
            for t in range(time_steps):
                total_standard_deviation[t] = math.sqrt(total_standard_deviation[t]) / sample_size
                target_mean_recovery[t] = mean_recovery[t][3] + mean_recovery[t][4]
                pdf_full[t] = sp.stats.norm.pdf(x_range, target_mean_recovery[t], total_standard_deviation[t])

            coeR = np.trapz(pdf_full, x_range)

            for t in range(time_steps):
                pdf_full[t] = pdf_full[t] / coeR[t]
                idx = int(len(x_range) * 95 / 100 + 1)
                irt[t] = np.trapz(pdf_full[t, idx:], x_range[idx:])

            pdf = np.zeros(time_steps)
            for t in range(1, time_steps - 1):
                pdf[t] = (irt[t + 1] - irt[t]) if irt[t + 1] - irt[t] > 0 else 0

            # Calculate truncated normal distribution and 75% & 95% percentile band
            # 75% percentile upper bound
            upper_bound75 = target_mean_recovery + [1.15 * i for i in total_standard_deviation]
            # 75% percentile lower bound
            lower_bound75 = target_mean_recovery - [1.15 * i for i in total_standard_deviation]
            # 95% percentile upper bound
            upper_bound95 = target_mean_recovery + [1.96 * i for i in total_standard_deviation]
            # 95% percentile lower bound
            lower_bound95 = target_mean_recovery - [1.96 * i for i in total_standard_deviation]

            for t in range(time_steps):
                coet = sp.stats.norm.cdf(0, target_mean_recovery[t], total_standard_deviation[t])
                coet2 = sp.stats.norm.cdf(1, target_mean_recovery[t], total_standard_deviation[t])

                if coet >= 0.000005 and 1 - coet2 < 0.00005:
                    coeAmp = 1 / (1 - coet)
                    lower_bound95[t] = sp.stats.norm.ppf(0.05 / coeAmp + coet, target_mean_recovery[t],
                                                         total_standard_deviation[t])
                    upper_bound95[t] = sp.stats.norm.ppf(0.95 / coeAmp + coet, target_mean_recovery[t],
                                                         total_standard_deviation[t])
                    lower_bound75[t] = sp.stats.norm.ppf(0.25 / coeAmp + coet, target_mean_recovery[t],
                                                         total_standard_deviation[t])
                    upper_bound75[t] = sp.stats.norm.ppf(0.75 / coeAmp + coet, target_mean_recovery[t],
                                                         total_standard_deviation[t])

                    pdf_full[t] = pdf_full[t] * coeAmp
                if coet < 0.00005 and 1 - coet2 >= 0.000005:
                    coeAmp = 1 / coet2
                    lower_bound95[t] = sp.stats.norm.ppf((coet2) - 0.95 / coeAmp, target_mean_recovery[t],
                                                         total_standard_deviation[t])
                    upper_bound95[t] = sp.stats.norm.ppf((coet2) - 0.05 / coeAmp, target_mean_recovery[t],
                                                         total_standard_deviation[t])
                    lower_bound75[t] = sp.stats.norm.ppf((coet2) - 0.75 / coeAmp, target_mean_recovery[t],
                                                         total_standard_deviation[t])
                    upper_bound75[t] = sp.stats.norm.ppf((coet2) - 0.25 / coeAmp, target_mean_recovery[t],
                                                         total_standard_deviation[t])

                    pdf_full[t] = pdf_full[t] * coeAmp
                if coet >= 0.000005 and 1 - coet2 >= 0.000005:
                    coeAmp = 1 / (coet2 - coet)
                    lower_bound95[t] = sp.stats.norm.ppf(0.05 / coeAmp + coet, target_mean_recovery[t],
                                                         total_standard_deviation[t])
                    upper_bound95[t] = sp.stats.norm.ppf(coet2 - 0.05 / coeAmp, target_mean_recovery[t],
                                                         total_standard_deviation[t])
                    lower_bound75[t] = sp.stats.norm.ppf(0.25 / coeAmp + coet, target_mean_recovery[t],
                                                         total_standard_deviation[t])
                    upper_bound75[t] = sp.stats.norm.ppf(coet2 - 0.25 / coeAmp, target_mean_recovery[t],
                                                         total_standard_deviation[t])

            # TODO: Confirm with PI if this can be removed. These conditions are never hit
            if time_steps > 100:
                for t in range(100):
                    if lower_bound75[t] < 0:
                        lower_bound75[t] = 1
                    if upper_bound75[t] < 0:
                        upper_bound75[t] = 1
                for t in range(100, time_steps):
                    if lower_bound95[t] < 0:
                        lower_bound95[t] = 1
                    if upper_bound95[t] < 0:
                        upper_bound95[t] = 1

            # END: Additional Code for uncertainty Analysis
            with open(output_base_name + '_portfolio-recovery.csv', 'w+', newline='') as output_file:
                spam_writer = csv.writer(output_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                spam_writer.writerow(['Week', 'Recovery_Percent_Func_Probability', '75P_Upper_Bound',
                                      '75P_Lower_Bound', '95P_Upper_Bound', '95P_Lower_Bound',
                                      'RecPercent_RE', 'RecPercent_RU', 'RecPercent_RO', 'RecPercent_BF',
                                      'RecPercent_FF', 'Probability_Density_Func'])
                for i in range(time_steps):
                    spam_writer.writerow([i + 1, mean_recovery_output[i], lower_bound75[i], upper_bound75[i],
                                          lower_bound95[i], upper_bound95[i]] + list(mean_recovery[i]) +
                                         [pdf[i]])
        else:
            with open(output_base_name + '_portfolio-recovery.csv', 'w+', newline='') as output_file:
                spam_writer = csv.writer(output_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                spam_writer.writerow(['Week', 'Recovery_Percent_Func_Probability', 'RecPercent_RE',
                                      'RecPercent_RU', 'RecPercent_RO', 'RecPercent_BF', 'RecPercent_FF'])
                for i in range(time_steps):
                    spam_writer.writerow([i + 1, mean_recovery_output[i]] + list(mean_recovery[i]))

        self.set_output_dataset("result", Dataset.from_file(output_base_name + '_portfolio-recovery.csv',
                                                            data_type=self.output_datasets["result"]["spec"]["type"]))

        print("INFO: Finished executing Building Portfolio Recovery Analysis")

        return True

    def calculate_transition_probability_matrix(self, time_steps, sample_buildings, repair_mean,
                                                occupancy_map,
                                                uncertainty, impeding_mean, impeding_std, building_damage,
                                                utility, utility2):
        sample_size = len(sample_buildings)
        total_mean = np.zeros((4, 4))
        total_var = np.zeros((4, 4))
        transition_probability = np.zeros((4, 4))
        state_probabilities = np.zeros((time_steps, 5))
        output = np.zeros(time_steps)
        temporary_correlation1 = np.zeros((time_steps, sample_size, 5))
        temporary_correlation2 = np.zeros((time_steps, sample_size, 5))
        mean_over_time = np.zeros((time_steps, sample_size))
        variance_over_time = np.zeros((time_steps, sample_size))
        recovery_fp = np.zeros((sample_size, time_steps))
        mean_recovery = np.zeros((time_steps, 5))
        print("Calculating transition probability matrix for each building..")
        for k in range(sample_size):
            # print("Calculating transition probability matrix for " + str(k))
            # The index for finance starts in 1 they are one off from the matrix
            finance_id = sample_buildings[k].finance - 1
            utility_id = sample_buildings[k].ep_pw_id
            for j in range(4):
                mean = impeding_mean[finance_id, j]
                std = impeding_std[finance_id, j] ** 2

                for i in range(j, 4):
                    mean += repair_mean[occupancy_map[sample_buildings[k].occupation_code]][i]
                    total_mean[j][i] = mean
                    std += 0.4 * repair_mean[occupancy_map[sample_buildings[k].occupation_code]][i] ** 2
                    total_var[j][i] = math.sqrt(std)

            for t in range(time_steps):
                for i in range(4):
                    for j in range(i, 4):
                        zeta = math.sqrt(math.log(1 + (total_var[i][j] / total_mean[i][j]) ** 2))
                        lambda_log = math.log(total_mean[i][j]) - 1 / 2 * zeta ** 2
                        transition_probability[i][j] = self.log_n_cdf(t + 1, lambda_log, zeta)
                # tpm = transition probability matrix
                tpm = np.matrix([[1 - transition_probability[0, 0],
                                  transition_probability[0, 0] - transition_probability[0, 1],
                                  transition_probability[0, 1] - transition_probability[0, 2],
                                  transition_probability[0, 2] - transition_probability[0, 3],
                                  transition_probability[0, 3]],
                                 [0.0, 1 - transition_probability[1, 1],
                                  transition_probability[1, 1] - transition_probability[1, 2],
                                  transition_probability[1, 2] - transition_probability[1, 3],
                                  transition_probability[1, 3]],
                                 [0.0, 0.0, 1 - transition_probability[2, 2],
                                  transition_probability[2, 2] - transition_probability[2, 3],
                                  transition_probability[2, 3]],
                                 [0.0, 0.0, 0.0, 1 - transition_probability[3, 3],
                                  transition_probability[3, 3]],
                                 [0.0, 0.0, 0.0, 0.0, 1.0]], dtype=float)
                # State Probability vector, pie(t) = initial vector * Transition Probability Matrix
                state_probabilities[t] = np.matmul(building_damage[k], tpm)
                if not uncertainty:
                    output[t] = state_probabilities[t, 3] + state_probabilities[t, 4]

                if uncertainty:
                    # Considering the effect of utility availability
                    # Utility Dependence Matrix
                    utility_matrix = np.matrix([[1, 0, 0, 0, 0],
                                                [0, 1, 0, 0, 0],
                                                [0, 0, 1, utility[utility_id][t], utility[utility_id][t]],
                                                [0, 0, 0, 1 - utility[utility_id][t],
                                                 utility2[utility_id][t]],
                                                [0, 0, 0, 0,
                                                 1 - utility[utility_id][t] - utility2[utility_id][t]]
                                                ], dtype=float)
                    updated_tpm = np.matmul(tpm, utility_matrix.transpose())
                    state_probabilities[t] = np.matmul(state_probabilities[t], utility_matrix.transpose())

                    # Calculation functionality statee indicator wheen j=4+5 Conditional mean
                    temporary_correlation1[t][k][0] = updated_tpm[0, 3]
                    temporary_correlation1[t][k][1] = updated_tpm[1, 3]
                    temporary_correlation1[t][k][2] = updated_tpm[2, 3]
                    temporary_correlation1[t][k][3] = updated_tpm[3, 3]
                    temporary_correlation1[t][k][4] = updated_tpm[4, 3]
                    temporary_correlation2[t][k][0] = updated_tpm[0, 4]
                    temporary_correlation2[t][k][1] = updated_tpm[1, 4]
                    temporary_correlation2[t][k][2] = updated_tpm[2, 4]
                    temporary_correlation2[t][k][3] = updated_tpm[3, 4]
                    temporary_correlation2[t][k][4] = updated_tpm[4, 4]
                    mean_over_time[t][k] = state_probabilities[t][3] + state_probabilities[t][4]
                    variance_over_time[t][k] = (state_probabilities[t][3] + state_probabilities[t][4]) * \
                                               (1 - (state_probabilities[t][3] + state_probabilities[t][4]))

            # Considering the effect of utility availability
            # Service Area ID of individual buildings
            # START: Code from only recovery analysis
            if not uncertainty:

                for i in range(len(state_probabilities)):
                    state_probabilities[i, 2] = state_probabilities[i, 2] + state_probabilities[i, 3] + \
                                                state_probabilities[i, 4] * (1 - utility[utility_id, i])
                    state_probabilities[i, 3] = state_probabilities[i, 3] * utility[utility_id, i]
                    state_probabilities[i, 4] = state_probabilities[i, 4] * utility[utility_id, i]

            # END: Code from only recovery analysis

            # Save functional probability (Best Line Functionality + Full functionality) for each building
            recovery_fp[k, :] = state_probabilities[:, 4] + state_probabilities[:, 3]

            # Aggregate state probability vector to portfolio level
            mean_recovery = mean_recovery + state_probabilities

        print("Transition probability matrix calculation complete.")

        return {"temporary_correlation1": temporary_correlation1,
                "temporary_correlation2": temporary_correlation2,
                "mean_over_time": mean_over_time, "variance_over_time": variance_over_time,
                "recovery_fp": recovery_fp,
                "mean_recovery": mean_recovery}

    # TODO: nS=10000 should be used line:301
    def calculate_sample_total(self, number_of_simulations, sample_size, building_damage, random_samples):
        sample_total = np.zeros((sample_size, number_of_simulations))
        for j in range(number_of_simulations):
            sample = np.zeros(sample_size)
            for i in range(sample_size):
                threshold = building_damage[i]
                if random_samples[j][i] <= threshold[0]:
                    sample[i] = 1
                elif random_samples[j][i] <= threshold[0] + threshold[1] and \
                        random_samples[j][i] >= threshold[0]:
                    sample[i] = 2
                elif threshold[0] + threshold[1] <= random_samples[j][i] <= \
                        threshold[0] + threshold[1] + threshold[2]:
                    sample[i] = 3
                elif threshold[0] + threshold[1] + threshold[2] <= random_samples[j][i] <= \
                        threshold[0] + threshold[1] + threshold[2] + threshold[3]:
                    sample[i] = 4
                else:
                    sample[i] = 5
                sample_total[i][j] = sample[i]
        return sample_total

    def calculate_std_of_mean_concurrent_future(self, function_name, parallelism, *args):

        output = []
        with concurrent.futures.ProcessPoolExecutor(max_workers=parallelism) as executor:
            for ret in executor.map(function_name, *args):
                output.extend(ret)
        return output

    def calculate_std_of_mean_bulk_input(self, time_steps, sample_size, number_of_simulations,
                                         variance_over_time, mean_over_time, temporary_correlation1,
                                         temporary_correlation2, sample_total):
        result = []
        for step in time_steps:
            result.append(
                self.calculate_std_of_mean(step, sample_size, number_of_simulations, variance_over_time,
                                           mean_over_time, temporary_correlation1, temporary_correlation2,
                                           sample_total))
        return result

    # calculating standard deviation of the mean recovery trajectory
    def calculate_std_of_mean(self, t, sample_size, number_of_simulations, variance_over_time, mean_over_time,
                              temporary_correlation1, temporary_correlation2, sample_total):

        print("Calculating std mean for week " + str(t))
        output = np.sum(variance_over_time[t])

        # Building i
        # start = timer()
        for i in range(sample_size - 1):
            # start = timer()
            # Building j
            for j in range(i + 1, sample_size):
                # starti = timer()
                expect1 = 0
                # Joint probability of initial functionality state P(S0i=k, S0j=l)
                joint_probability = self.joint_probability_calculation(sample_total[i], sample_total[j],
                                                                       number_of_simulations)

                # Functionality State k
                for k in range(5):
                    # Functionality State l
                    for l in range(5):
                        expect1 += joint_probability[k][l] * (temporary_correlation1[t][i][k] *
                                                              temporary_correlation1[t][j][l]
                                                              + temporary_correlation1[t][i][k] *
                                                              temporary_correlation1[t][j][l]
                                                              + temporary_correlation2[t][i][k] *
                                                              temporary_correlation1[t][j][l]
                                                              + temporary_correlation2[t][i][k] *
                                                              temporary_correlation2[t][j][l])
                expect2 = mean_over_time[t][i] * mean_over_time[t][j]

                if variance_over_time[t][i] > 0 and variance_over_time[t][j] > 0 and expect1 - expect2 > 0:
                    covariance = expect1 - expect2
                    output += 2 * covariance

        return output

    # TODO: Review
    def calculate_delay_time(self, res_buildings, finance):
        """This function calculates the delay time given an initial functionality state and a financing resource.

        Args:
            res_buildings (int): Initial Damage State 1: Slight Damage.
            finance (int): Financing types, Can be a number from 1 to 5:

                1. Insurance
                2. SBA Business Loan
                3. Private
                4. Savings
                5. Not covered

        Returns:
            np.Array: Impeding. It takes into account each factor that alters the recovery time.
            Each item in the temporary array represents:

            1. impeding[0]: Initial Damage State
            2. impeding[1]: Engineering Mobilization
            3. impeding[2]: Insurance
            4. impeding[3]: Contractor Mobilization
            5. impeding[4]: Obtain permits

        """
        impeding = np.zeros(5)

        if res_buildings == 3:
            impeding[0] = 0
            impeding[1] = np.random.lognormal(np.log(0.5), 0.4)
            impeding[2] = 0
            impeding[3] = np.random.lognormal(np.log(3), 0.6)

        elif res_buildings == 1 or res_buildings == 2:
            impeding[0] = np.random.lognormal(np.log(1), 0.54)
            if res_buildings == 1:
                impeding[1] = np.random.lognormal(np.log(15), 0.32)
                impeding[3] = np.random.lognormal(np.log(12), 0.38)

            else:
                impeding[1] = np.random.lognormal(np.log(3), 0.4)
                impeding[3] = np.random.lognormal(np.log(6), 0.6)

            if finance == 0:
                impeding[2] = np.random.lognormal(np.log(3), 1.11)
            elif finance == 1:
                impeding[2] = np.random.lognormal(np.log(10), 0.57)
            elif finance == 2:
                impeding[2] = np.random.lognormal(np.log(7), 0.68)
            elif finance == 3:
                impeding[2] = 0
            else:
                impeding[2] = np.random.lognormal(np.log(15), 0.65)

        else:
            impeding[0] = np.random.lognormal(np.log(4), 0.54, 1)
            impeding[1] = np.random.lognormal(np.log(15), 0.32, 1)

            if finance == 0:
                impeding[2] = np.random.lognormal(np.log(6), 1.11)
            elif finance == 1:
                impeding[2] = np.random.lognormal(np.log(30), 0.57)
            elif finance == 2:
                impeding[2] = np.random.lognormal(np.log(15), 0.68)
            elif finance == 3:
                impeding[2] = 0
            else:
                impeding[2] = np.random.lognormal(np.log(40), 0.65)

            impeding[3] = np.random.lognormal(np.log(12), 0.38)

        if res_buildings == 2 or res_buildings == 3:
            impeding[4] = 0
        else:
            impeding[4] = np.random.lognormal(np.log(6), 0.32)

        return impeding[0] + max(impeding[1:4]) + impeding[4]

    # TODO: Improve readability
    def joint_probability_calculation(self, sample_i, sample_j, number_of_samples):
        output = np.zeros((5, 5))
        for n in range(number_of_samples):
            if sample_i[n] == 1 and sample_j[n] == 1:
                output[0][0] += 1.0 / number_of_samples
            elif sample_i[n] == 1 and sample_j[n] == 2:
                output[0][1] += 1.0 / number_of_samples
            elif sample_i[n] == 1 and sample_j[n] == 3:
                output[0][2] += 1.0 / number_of_samples
            elif sample_i[n] == 1 and sample_j[n] == 4:
                output[0][3] += 1.0 / number_of_samples
            elif sample_i[n] == 1 and sample_j[n] == 5:
                output[0][4] += 1.0 / number_of_samples
            if sample_i[n] == 2 and sample_j[n] == 1:
                output[1][0] += 1.0 / number_of_samples
            elif sample_i[n] == 2 and sample_j[n] == 2:
                output[1][1] += 1.0 / number_of_samples
            elif sample_i[n] == 2 and sample_j[n] == 3:
                output[1][2] += 1.0 / number_of_samples
            elif sample_i[n] == 2 and sample_j[n] == 4:
                output[1][3] += 1.0 / number_of_samples
            elif sample_i[n] == 2 and sample_j[n] == 5:
                output[1][4] += 1.0 / number_of_samples
            if sample_i[n] == 3 and sample_j[n] == 1:
                output[2][0] += 1.0 / number_of_samples
            elif sample_i[n] == 3 and sample_j[n] == 2:
                output[2][1] += 1.0 / number_of_samples
            elif sample_i[n] == 3 and sample_j[n] == 3:
                output[2][2] += 1.0 / number_of_samples
            elif sample_i[n] == 3 and sample_j[n] == 4:
                output[2][3] += 1.0 / number_of_samples
            elif sample_i[n] == 3 and sample_j[n] == 5:
                output[2][4] += 1.0 / number_of_samples
            if sample_i[n] == 4 and sample_j[n] == 1:
                output[3][0] += 1.0 / number_of_samples
            elif sample_i[n] == 4 and sample_j[n] == 2:
                output[3][1] += 1.0 / number_of_samples
            elif sample_i[n] == 4 and sample_j[n] == 3:
                output[3][2] += 1.0 / number_of_samples
            elif sample_i[n] == 4 and sample_j[n] == 4:
                output[3][3] += 1.0 / number_of_samples
            elif sample_i[n] == 4 and sample_j[n] == 5:
                output[3][4] += 1.0 / number_of_samples
            if sample_i[n] == 5 and sample_j[n] == 1:
                output[4][0] += 1.0 / number_of_samples
            elif sample_i[n] == 5 and sample_j[n] == 2:
                output[4][1] += 1.0 / number_of_samples
            elif sample_i[n] == 5 and sample_j[n] == 3:
                output[4][2] += 1.0 / number_of_samples
            elif sample_i[n] == 5 and sample_j[n] == 4:
                output[4][3] += 1.0 / number_of_samples
            elif sample_i[n] == 5 and sample_j[n] == 5:
                output[4][4] += 1.0 / number_of_samples

        return output

    def log_n_cdf(self, x, mean, sig):
        if x < 0:
            cdf = 0
        elif np.isposinf(x):
            cdf = 1
        else:
            z = (np.log(x) - mean) / float(sig)
            cdf = 0.5 * (math.erfc(-z / np.sqrt(2)))

        return cdf



# Copyright (c) 2022 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import numpy as np
import pandas as pd
from scipy.stats import poisson, bernoulli

from pyincore import BaseAnalysis


class PipelineFunctionality(BaseAnalysis):
    """
    This analysis computes pipeline functionality using repair rate calculations from pipeline damage analysis
    (earthquake).
    The computation operates by computing Monte Carlo samples derived from Poisson sample deviates from the damage
    analysis as input to Bernoulli experiments, later used to determine average functionality.
    The output of the computation is the average pipeline functionality.

    Contributors
        | Science: Neetesh Sharma, Armin Tabandeh, Paolo Gardoni
        | Implementation: Neetesh Sharma, Chen Wang, and NCSA IN-CORE Dev Team

    Related publications
        Sharma, N., Tabandeh, A., & Gardoni, P. (2019). Regional resilience analysis: A multi-scale approach to model
        the recovery of interdependent infrastructure. In P. Gardoni (Ed.), Handbook of sustainable and resilient
        infrastructure (pp. 521–544). New York, NY: Routledge.
        Sharma, N., Tabandeh, A., & Gardoni, P. (2020). Regional resilience analysis: A multi-scale approach to
        optimize the resilience of interdependent infrastructure. Computer‐Aided Civil and Infrastructure Engineering,
        35(12), 1315-1330.
        Sharma, N., & Gardoni, P. (2022). Mathematical modeling of interdependent infrastructure: An object-oriented
        approach for generalized network-system analysis. Reliability Engineering & System Safety, 217, 108042.

    Args:
    incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        super(PipelineFunctionality, self).__init__(incore_client)

    def run(self):
        pipeline_dmg_df = self.get_input_dataset("pipeline_repair_rate_damage").get_dataframe_from_csv()

        num_samples = self.get_parameter("num_samples")

        (fs_results, fp_results) = self.pipeline_functionality(pipeline_dmg_df, num_samples)
        self.set_result_csv_data("sample_failure_state",
                                 fs_results, name=self.get_parameter("result_name") + "_failure_state",
                                 source="dataframe")
        self.set_result_csv_data("failure_probability",
                                 fp_results,
                                 name=self.get_parameter("result_name") + "_failure_probability",
                                 source="dataframe")

        return True

    def pipeline_functionality(self, pipeline_dmg_df, num_samples):
        """Run pipeline functionality analysis for multiple pipelines.

        Args:
            pipeline_dmg_df (dataframe): dataframe of pipeline damage values and other data/metadata
            num_samples (int): number of samples
        Returns:
            fs_results (list): A list of dictionary with id/guid and failure state for N samples
            fp_results (list): A list dictionary with failure probability and other data/metadata.

        """

        pipeline_dmg_df['pgv_pf'] = 1 - poisson.pmf(0, pipeline_dmg_df.loc[:, 'numpgvrpr'].values)

        # todo there is more efficient pandas manipulation
        sampcols = ['s' + samp for samp in np.arange(num_samples).astype(str)]

        fs_results = pd.DataFrame(
            bernoulli.rvs(1 - pipeline_dmg_df.loc[:, 'pgv_pf'].values, size=(num_samples, pipeline_dmg_df.shape[0])).T,
            index=pipeline_dmg_df.guid.values, columns=sampcols)
        fp_results = fs_results.copy(deep=True)

        # calculate sample failure
        # concatenate all columns into one failure column
        fs_results['failure'] = fs_results.astype(str).apply(','.join, axis=1)
        fs_results = fs_results.filter(['failure'])
        # set guid column
        fs_results.reset_index(inplace=True)
        fs_results = fs_results.rename(columns={'index': 'guid'})

        # calculate failure probability
        # count of 0s divided by sample size
        fp_results["failure_probability"] = (num_samples - fp_results.sum(axis=1).astype(int)) / num_samples
        fp_results = fp_results.filter(['failure_probability'])
        # set guid column
        fp_results.reset_index(inplace=True)
        fp_results = fp_results.rename(columns={'index': 'guid'})

        return fs_results, fp_results

    def get_spec(self):
        """Get specifications of the pipeline functionality analysis.

        Returns:
            obj: A JSON object of specifications of the pipeline functionality analysis.

        """
        return {
            'name': 'pipeline-functionality',
            'description': 'buried pipeline functionality analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'num_samples',
                    'required': True,
                    'description': 'Number of MC samples',
                    'type': int
                },
            ],
            'input_datasets': [
                {
                    'id': 'pipeline_repair_rate_damage',
                    'required': True,
                    'description': 'Output of pipeline damage repair rate analysis',
                    'type': ['ergo:pipelineDamageVer3'],
                },
            ],
            'output_datasets': [
                {
                    'id': 'failure_probability',
                    'description': 'CSV file of failure probability',
                    'type': 'incore:failureProbability'
                },
                {
                    'id': 'sample_failure_state',
                    'description': 'CSV file of failure state for each sample',
                    'type': 'incore:sampleFailureState'
                },
            ]
        }



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import numpy as np
import pandas as pd

from pyincore import BaseAnalysis


class BuildingFunctionality(BaseAnalysis):
    """The building functionality analysis can be used to calculate building functionality probabilities considering
    two situations: buildings are in at least a damage state 2 or greater or buildings are not damaged but electric
    power is not available to the building. Whether buildings can receive electrical power is assumed to depend on
    the interdependency between buildings and substations, and between buildings and poles in close proximity.
    If both the nearest pole to the building and the substation where buildings belong to its service area are
    functional, buildings are considered to be able to receive electric power.

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        super().__init__(incore_client)

    def get_spec(self):
        """Get specifications of the building functionality analysis.

        Returns:
            obj: A JSON object of specifications of the building functionality analysis.

        """
        return {
            'name': 'functionality_probability',
            'description': 'calculate the functionality probability of each building',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': False,
                    'description': 'result dataset name',
                    'type': str
                }
            ],
            'input_datasets': [
                {
                    'id': 'building_damage_mcs_samples',
                    'required': True,
                    'description': 'building damage samples',
                    'type': ['incore:sampleFailureState'],
                },
                {
                    'id': 'substations_damage_mcs_samples',
                    'required': False,
                    'description': 'substations damage samples',
                    'type': ['incore:sampleFailureState'],
                },
                {
                    'id': 'poles_damage_mcs_samples',
                    'required': False,
                    'description': 'poles damage samples',
                    'type': ['incore:sampleFailureState'],
                },
                {
                    'id': 'interdependency_dictionary',
                    'required': False,
                    'description': 'JSON file of interdependency between buildings and substations and poles',
                    'type': ['incore:buildingInterdependencyDict'],
                },
            ],
            'output_datasets': [
                {
                    'id': 'functionality_samples',
                    'description': 'CSV file of functionality samples',
                    'type': 'incore:funcSample'
                },
                {
                    'id': 'functionality_probability',
                    'description': 'CSV file of functionality probability',
                    'type': 'incore:funcProbability'
                }
            ]
        }

    def run(self):
        """Executes building functionality analysis"""

        # enable index on "guid" column
        buildings_df = self.get_input_dataset("building_damage_mcs_samples").get_dataframe_from_csv().set_index("guid")

        interdependency_dataset = self.get_input_dataset("interdependency_dictionary")
        if interdependency_dataset is not None:
            interdependency_dict = interdependency_dataset.get_json_reader()
        else:
            interdependency_dict = None

        substations_dataset = self.get_input_dataset("substations_damage_mcs_samples")
        if substations_dataset is not None:
            substations_df = substations_dataset.get_dataframe_from_csv().set_index("guid")
        else:
            substations_df = None

        poles_dataset = self.get_input_dataset("poles_damage_mcs_samples")
        if poles_dataset is not None:
            poles_df = poles_dataset.get_dataframe_from_csv().set_index("guid")
        else:
            poles_df = None

        if (poles_dataset is not None or substations_dataset is not None) and interdependency_dataset is None:
            raise ValueError("Please provide interdependency table if pole or substation damage is "
                             "considered in the building functionality calculation.")

        functionality_probabilities = []
        functionality_samples = []
        for building_guid in buildings_df.index:
            building_guid, sample, probability = self.functionality(building_guid, buildings_df, substations_df,
                                                                    poles_df,
                                                                    interdependency_dict)
            functionality_probabilities.append([building_guid, probability])
            functionality_samples.append([building_guid, sample])

        fp_results = pd.DataFrame(functionality_probabilities, columns=['building_guid', 'probability'])
        fs_results = pd.DataFrame(functionality_samples, columns=['building_guid', 'samples'])

        self.set_result_csv_data("functionality_probability",
                                 fp_results,
                                 name=self.get_parameter("result_name") + "_functionality_probability",
                                 source='dataframe')

        self.set_result_csv_data("functionality_samples",
                                 fs_results,
                                 name=self.get_parameter("result_name") + "_functionality_samples",
                                 source='dataframe')

        return True

    def functionality(self, building_guid, buildings, substations, poles, interdependency):
        """

        Args:
            building_guid (str): A building defined by its guid.
            buildings (pd.DataFrame): A list of buildings.
            substations (pd.DataFrame): A list of substations.
            poles (pd.DataFrame): A list of poles.
            interdependency (dict): An interdependency between buildings and substations and poles.

        Returns:
            str: A building guid.
            str: A functionality sample that is a string of "0,0,1...".
            str: A probability [0,1] of building being functional.

        """

        building_mc_samples = buildings.loc[building_guid]
        building_list = []
        try:
            building_list = building_mc_samples["failure"].split(",")
        except IndexError:
            print("error with buildings")
            print(building_mc_samples)
            return {building_guid: -1}

        # if building is defined in the interdependency lookup table
        if interdependency is not None:

            if building_guid in interdependency.keys():
                if substations is not None:
                    substations_mc_samples = substations.loc[interdependency[building_guid]["substations_guid"]]
                    substation_list = []
                    try:
                        substation_list = substations_mc_samples["failure"].split(",")
                    except IndexError:
                        print("error with substations")
                        print(interdependency[building_guid]["substations_guid"])
                        return {building_guid: -1}
                else:
                    substation_list = None

                if poles is not None:
                    poles_mc_samples = poles.loc[interdependency[building_guid]["poles_guid"]]
                    pole_list = []
                    try:
                        pole_list = poles_mc_samples["failure"].split(",")
                    except IndexError:
                        print("error with poles")
                        print(interdependency[building_guid]["poles_guid"])
                        return {building_guid: -1}
                else:
                    pole_list = None

                if substation_list is not None and pole_list is not None:
                    functionality_samples = [BuildingFunctionality._calc_functionality_samples(building_sample,
                                                                                               substation_sample,
                                                                                               pole_sample)
                                             for building_sample, substation_sample, pole_sample in
                                             zip(building_list, substation_list, pole_list)]
                elif substation_list is not None:
                    functionality_samples = [BuildingFunctionality._calc_functionality_samples(building_sample,
                                                                                               substation_sample,
                                                                                               None)
                                             for building_sample, substation_sample in
                                             zip(building_list, substation_list)]
                elif pole_list is not None:
                    functionality_samples = [BuildingFunctionality._calc_functionality_samples(building_sample,
                                                                                               None,
                                                                                               pole_sample)
                                             for building_sample, pole_sample in
                                             zip(building_list, pole_list)]
                else:
                    functionality_samples = [BuildingFunctionality._calc_functionality_samples(building_sample,
                                                                                               None,
                                                                                               None)
                                             for building_sample in building_list]
                probability = BuildingFunctionality._calc_functionality_probability(functionality_samples)
                return building_guid, ",".join([str(sample) for sample in functionality_samples]), probability

            else:
                return building_guid, "NA", "NA"

        # else if only building MC failure is available
        else:
            functionality_samples = [BuildingFunctionality._calc_functionality_samples(building_sample) for
                                     building_sample in building_list]
            probability = BuildingFunctionality._calc_functionality_probability(functionality_samples)
            return building_guid, ",".join([str(sample) for sample in functionality_samples]), probability

    @staticmethod
    def _calc_functionality_samples(building_sample, substation_sample=None, pole_sample=None):
        """ This function is subject to change. For now, buildings have a 1-to-1 relationship with
        substations and poles, so it suffices to check that the poles and substations are up.

        Args:
            building_sample (str): Monte Carlo samples of building functionality.
            substation_sample (str|None): Monte Carlo samples of substation functionality.
            pole_sample (str|None): Monte Carlo samples of pole functionality.

        Returns:
            int: 1 if building is functional, 0 otherwise

        """
        if building_sample == "1" and (substation_sample == "1" or substation_sample is None) \
                and (pole_sample == "1" or pole_sample is None):
            return 1
        else:
            return 0

    @staticmethod
    def _calc_functionality_probability(functionality_samples):
        functionality_sum = np.sum(functionality_samples)
        num_samples = len(functionality_samples)
        probability = 0.0
        if functionality_sum > 0:
            probability = (functionality_sum / num_samples)

        return probability
